import io
from contextlib import contextmanager

import numpy as np

import eccodes

from ._bufr import BUFRMessage  # noqa

_TYPES_MAP = {
    "float": float,
    "int": int,
    "str": str,
}


@contextmanager
def raise_keyerror(name):
    """Make operations on a key raise a KeyError if not found"""
    try:
        yield
    except (eccodes.KeyValueNotFoundError, eccodes.FunctionNotImplementedError):
        raise KeyError(name)


class Message:
    def __init__(self, handle):
        self._handle = handle

    def __del__(self):
        try:
            eccodes.codes_release(self._handle)
        except Exception:
            pass

    def copy(self):
        """Create a copy of the current message"""
        return self.__class__(eccodes.codes_clone(self._handle))

    def __copy__(self):
        return self.copy()

    def _get(self, name, ktype=None):
        name, sep, stype = name.partition(":")
        if sep and ktype is None:
            try:
                ktype = _TYPES_MAP[stype]
            except KeyError:
                raise ValueError(f"Unknown key type {stype!r}")
        with raise_keyerror(name):
            if eccodes.codes_is_missing(self._handle, name):
                raise KeyError(name)
            if eccodes.codes_get_size(self._handle, name) > 1:
                return eccodes.codes_get_array(self._handle, name, ktype=ktype)
            return eccodes.codes_get(self._handle, name, ktype=ktype)

    def get(self, name, default=None, ktype=None):
        """Get the value of a key

        Parameters
        ----------
        name: str
            Name of the key. Can be suffixed with ":str", ":int", or ":float" to
            request a specific type.
        default: any, optional
            Value if the key is not Found, or ``None`` if not specified.
        ktype: type
            Request a specific type for the value. Overrides the suffix in ``name``"""
        try:
            return self._get(name, ktype=ktype)
        except KeyError:
            return default

    def set(self, *args, check_values: bool = True):
        """If two arguments are given, assumes this takes form of a single key
        value pair and sets the value of the given key. If a dictionary is passed in,
        then sets the values of all keys in the dictionary. Note, ordering
        of the keys is important. Finally, by default, checks if values
        have been set correctly

        Raises
        ------
        TypeError
            If arguments do not take one of the two expected forms
        KeyError
            If the key does not exist
        ValueError
            If the set value of one of the keys is not the expected value
        """
        if isinstance(args[0], str) and len(args) == 2:
            key_values = {args[0]: args[1]}
        elif isinstance(args[0], dict):
            key_values = args[0]
        else:
            raise TypeError(
                "Unsupported argument type. Expects two arguments consisting \
            of key and value pair, or a dictionary of key-value pairs"
            )

        for name, value in key_values.items():
            with raise_keyerror(name):
                if np.ndim(value) > 0:
                    eccodes.codes_set_array(self._handle, name, value)
                else:
                    eccodes.codes_set(self._handle, name, value)

        if check_values:
            # Check values just set
            for name, value in key_values.items():
                if type(value) in _TYPES_MAP.values():
                    saved_value = self.get(f"{name}:{type(value).__name__}")
                else:
                    saved_value = self.get(name)
                if not np.all(saved_value == value):
                    raise ValueError(
                        f"Unexpected retrieved value {saved_value} for key {name}. Expected {value}"
                    )

    def get_array(self, name):
        """Get the value of the given key as an array

        Raises
        ------
        KeyError
            If the key is not set
        """
        with raise_keyerror(name):
            return eccodes.codes_get_array(self._handle, name)

    def get_size(self, name):
        """Get the size of the given key

        Raises
        ------
        KeyError
            If the key is not set
        """
        with raise_keyerror(name):
            return eccodes.codes_get_size(self._handle, name)

    def get_data_points(self):
        raise NotImplementedError

    def is_missing(self, name):
        """Check whether the key is set to a missing value

        Raises
        ------
        KeyError
            If the key is not set
        """
        with raise_keyerror(name):
            return bool(eccodes.codes_is_missing(self._handle, name))

    def set_array(self, name, value):
        """Set the value of the given key

        Raises
        ------
        KeyError
            If the key does not exist
        """
        with raise_keyerror(name):
            return eccodes.codes_set_array(self._handle, name, value)

    def set_missing(self, name):
        """Set the given key as missing

        Raises
        ------
        KeyError
            If the key does not exist
        """
        with raise_keyerror(name):
            return eccodes.codes_set_missing(self._handle, name)

    def __getitem__(self, name):
        return self._get(name)

    def __setitem__(self, name, value):
        self.set(name, value)

    def __contains__(self, name):
        return bool(eccodes.codes_is_defined(self._handle, name))

    class _KeyIterator:
        def __init__(self, message, namespace=None, iter_keys=True, iter_values=False):
            self._message = message
            self._iterator = eccodes.codes_keys_iterator_new(message._handle, namespace)
            self._iter_keys = iter_keys
            self._iter_values = iter_values

        def __del__(self):
            try:
                eccodes.codes_keys_iterator_delete(self._iterator)
            except Exception:
                pass

        def __iter__(self):
            return self

        def __next__(self):
            while True:
                if not eccodes.codes_keys_iterator_next(self._iterator):
                    raise StopIteration
                if not self._iter_keys and not self._iter_values:
                    return
                key = eccodes.codes_keys_iterator_get_name(self._iterator)
                if self._message.is_missing(key):
                    continue
                if self._iter_keys and not self._iter_values:
                    return key
                value = self._message.get(key) if self._iter_values else None
                if not self._iter_keys:
                    return value
                return key, value

    def __iter__(self):
        return self._KeyIterator(self)

    def keys(self, namespace=None):
        """Iterate over all the available keys"""
        return self._KeyIterator(self, namespace, iter_keys=True, iter_values=False)

    def values(self, namespace=None):
        """Iterate over the values of all the available keys"""
        return self._KeyIterator(self, namespace, iter_keys=False, iter_values=True)

    def items(self, namespace=None):
        """Iterate over all the available key-value pairs"""
        return self._KeyIterator(self, namespace, iter_keys=True, iter_values=True)

    def dump(self):
        """Print out a textual representation of the message"""
        eccodes.codes_dump(self._handle)

    def write_to(self, fileobj):
        """Write the message to a file object"""
        assert isinstance(fileobj, io.IOBase)
        eccodes.codes_write(self._handle, fileobj)

    def get_buffer(self):
        """Return a buffer containing the encoded message"""
        return eccodes.codes_get_message(self._handle)


class GRIBMessage(Message):
    def __init__(self, handle):
        super().__init__(handle)
        self._data = None

    @property
    def data(self):
        """Return the array of values"""
        if self._data is None:
            self._data = self._get("values")
        return self._data

    def get_data_points(self):
        """Get the list of ``(lat, lon, value)`` data points"""
        return eccodes.codes_grib_get_data(self._handle)

    @classmethod
    def from_samples(cls, name):
        """Create a message from a sample"""
        return cls(eccodes.codes_grib_new_from_samples(name))



# Copyright 2022- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import re
from copy import copy
from itertools import repeat

from .common import *
from .helpers import RaggedArray, SingletonDict
from .tables import Code, Element

# flake8: noqa: F405
#   ruff: noqa: F403


INT_UNITS = re.compile(r"%|.*CODE TABLE.*|.*FLAG TABLE.*|Numeric|a|d|h|min|mon|s")


@dataclass
class Node:
    parent: Any = field(repr=False)
    ordinal: int = 0  # [1]
    level: int = 0  # [2]
    depth: int = 0  # [3]
    children: List = field(default_factory=list)
    starts: Dict[MultiIndex, Counter] = field(default_factory=dict)
    slices: Dict[MultiIndex, slice] = field(default_factory=dict)
    max_levels: Dict[str, int] = field(default_factory=dict)  # [4]

    def __post_init__(self):
        if self.parent:
            self.depth = self.parent.depth + 1

    # [1] Node's (0-based) ordinal number in its parent's list of children.
    #
    # [2] Replication level: leaf node(s) of the outer-most replication have
    #     level=0, in the next inner replication level=1, etc. This level
    #     also corresponds to the `ndim` of the RaggedArray of replication
    #     factors.
    #
    # [3] Distance from the root node. Note that this is not the same as the
    #     (replication) level!
    #
    # [4] A dictionary of the max. (replication) level of each key accessible
    #     from this node. For WrapperNodes this includes keys from all children
    #     subnodes, and their subnotes, transitively. This is used in DataView
    #     objects to infer shapes of the respective array views.


@dataclass
class WrapperNode(Node):
    counts: Dict[MultiIndex, Counter] = field(default_factory=dict)


@dataclass
class ReplicationNode(Node):
    counts: Dict[MultiIndex, Counter] = field(default_factory=dict)
    element: Element = field(default_factory=Element)
    factors: RaggedArray = field(default_factory=RaggedArray.empty)
    reset: bool = True

    def __post_init__(self):
        super().__post_init__()
        self.factors = RaggedArray.empty(self.level)


@dataclass
class LeafNode(Node):
    keys: List[Key] = field(default_factory=list)
    counts: SingletonDict = field(default_factory=SingletonDict)


@dataclass
class AssociationNode(Node):
    operator: int = 0
    reuse_bitmap: bool = False
    for_reuse: bool = False
    bitmap_size: Optional[int] = None
    center: Element = field(default_factory=Element)
    generating_application: Element = field(default_factory=Element)
    code_table: int = 0
    value_factor: Optional[Code] = None
    value_size: Optional[int] = None
    element: Element = field(default_factory=Element)
    bitmap: Optional[NDArray] = None
    bitmap_end: int = 0
    keys: List[Key] = field(default_factory=list)
    counts: Dict[MultiIndex, Counter] = field(default_factory=dict)


def build_tree(coder):
    tables = coder.get_tables()
    operators = {4: []}  # key: Code.X, value: a stack of Code.Y values (aka operands)

    def parse(parent, codes, level=0):
        keys = []
        children = []
        change_scale = None
        change_width = None
        while True:
            if not codes:
                if keys:
                    leaf = LeafNode(parent, len(children), level, keys=keys)
                    children.append(leaf)
                break
            next = codes.pop(0)
            # Sequence
            if next.F == 3:
                sequence_codes = list(tables.expand_codes(next, recursive=False))
                codes = sequence_codes + codes
            # Operator: Cancel reuse of bitmap
            elif next == 223255:
                next = codes.pop(0)
            # Operator: Quality information follows
            elif next in (222000, 223000, 224000, 225000):
                if keys:
                    leaf = LeafNode(parent, len(children), level, keys=keys)
                    children.append(leaf)
                    keys = []
                node = AssociationNode(parent, len(children), level, operator=next)
                next = codes.pop(0)
                assert next in [236000, 237000] or next.F == 1
                # Operator: Define bitmap for possible reuse
                if next == 236000:
                    node.for_reuse = True
                    next = codes.pop(0)
                # Process bitmap
                if next != 237000:
                    if next.F == 1:  # replication
                        assert next.X == 1  # replicate only one element
                        if next.Y == 0:  # delayed replication
                            next = codes.pop(0)
                            assert next.FX == (0, 31)  # delayed replication factor
                            element = tables.elements[next]
                            node.keys.append(
                                Key(
                                    element.name,
                                    element=element,
                                    flags=CODED | SCALAR | FACTOR | READ_ONLY,
                                )
                            )
                        else:  # fixed replication
                            node.bitmap_size = next.Y
                        next = codes.pop(0)
                    else:  # no replication
                        node.bitmap_size = 1
                    assert next == 31031  # dataPresentIndicator
                    element = tables.elements[next]
                    # TODO node.keys.append(Key(element.name, element=element, flags=CODED|BITMAP|READ_ONLY))
                # Operator: Reuse bitmap
                else:
                    node.reuse_bitmap = True
                # The next element is expected to be 'centre'
                next = codes.pop(0)
                element = tables.elements[next]
                assert element.name == "centre"
                node.center = element
                node.keys.append(Key(element.name, element=element, flags=CODED))
                # The next elements are expected to be 'generatingApplication'
                # Note that the generalApplication element can show up in
                # the local table too (e.g. 001201 from local/1/98/0).
                next = codes.pop(0)
                element = tables.elements[next]
                assert element.name == "generatingApplication"
                node.generating_application = element
                node.keys.append(Key(element.name, element=element, flags=CODED))
                next = codes.pop(0)
                # Class 8 element (optional)
                if next.X == 8:
                    node.code_table = next
                    element = tables.elements[next]
                    node.keys.append(Key(element.name, element=element, flags=CODED))
                    next = codes.pop(0)
                # Process the actual quality element (e.g. percentConfidence)
                if next.F == 1:  # replication
                    assert next.X == 1  # replicate only one element
                    if next.Y == 0:  # delayed replication
                        next = codes.pop(0)
                        node.value_factor = next
                        assert next.FX == (0, 31)  # delayed replication factor
                        element = tables.elements[next]
                        node.keys.append(
                            Key(
                                element.name,
                                element=element,
                                flags=CODED | SCALAR | FACTOR | READ_ONLY,
                            )
                        )
                    else:  # fixed replication
                        node.value_size = next.Y
                    next = codes.pop(0)
                else:  # no replication
                    pass
                if node.operator == 222000:
                    assert next.FX == (0, 33)  # expecting Class 33 element
                    node.element = tables.elements[next]
                else:
                    if node.operator == 223000:
                        assert next == 223255
                        name = "substitutedValue"
                    elif node.operator == 224000:
                        assert next == 224255
                        name = "firstOrderStatisticalValue"
                    elif node.operator == 225000:
                        assert next == 225255
                        name = "differenceStatisticalValue"
                    node.element = Element(next, name, "", 0, 0, 0)
                # When there is no prior replication, the associated element can
                # appear two or more times (aka "inlined" replication).
                if not node.value_factor and node.value_size is None:
                    node.value_size = 1
                    while codes and codes[0] == node.element.code:
                        next = codes.pop(0)
                        node.value_size += 1
                children.append(node)
            # Operator
            elif next.F == 2:
                # Operator: Change width
                if next.X == 1:
                    if next.Y > 0:
                        change_width = next.Y - 128
                    else:
                        change_width = None
                # Operator: Change scale
                if next.X == 2:
                    if next.Y > 0:
                        change_scale = next.Y - 128
                    else:
                        change_scale = None
                # Operator: Add associated field
                elif next.X == 4:
                    if next.Y > 0:
                        if codes[0] != 31021:  # Associated field significance
                            # TODO: malformed message: raise warning
                            operators[4].append(8)
                        else:
                            next = codes.pop(0)
                            operators[4].append(next.Y)
                    else:
                        operators[4].pop()
                # Operator: Signify data width for the following local descriptor
                elif next.X == 6:
                    if codes[0] not in tables.elements:
                        codes.pop(0)
            # Replication
            elif next.F == 1:
                span = next.X
                # Delayed replication
                if next.Y == 0:
                    next = codes.pop(0)
                    assert next.FX == (0, 31) and next.Y in (
                        0,
                        1,
                        2,
                        11,
                        12,
                    )  # delayed repl. factor
                    element = tables.elements[next]
                    key = Key(
                        element.name,
                        element=element,
                        flags=CODED | SCALAR | FACTOR | READ_ONLY,
                    )
                    keys.append(key)
                    leaf = LeafNode(parent, len(children), level, keys=keys)
                    children.append(leaf)
                    wrapper = WrapperNode(parent, len(children), level)
                    replication = ReplicationNode(wrapper, 0, level, element=element)
                    wrapper.children.append(replication)
                    parse(replication, codes[:span], level + 1)
                    children.append(wrapper)
                    codes = codes[span:]
                    keys = []
                # Fixed replication
                else:
                    assert next.X <= len(codes)
                    codes = codes[: next.X] * next.Y + codes[next.X :]
            # Element
            elif next.F == 0:
                element = tables.elements[next]
                if change_width or change_scale:
                    element = copy(element)
                    if change_width:
                        element.width += change_width
                    if change_scale:
                        element.scale += change_scale
                key = Key(element.name, element=element)
                keys.append(key)
                # Append associated field keys
                for Y in operators[4]:
                    name = f"{element.name}->associatedField"
                    key = Key(
                        name, element=Element(0, name, "associated units", 0, 0, 2)
                    )
                    keys.append(key)
                    name = f"{name}->associatedFieldSignificance"
                    key = Key(name, element=tables.elements[31021])
                    keys.append(key)
            else:
                assert False
        parent.children = children

    # Parse (unexpanded) descriptors

    codes = coder.get("unexpandedDescriptors")
    codes = [Code(c) for c in codes.tolist()]
    subset_count = coder.get("numberOfSubsets")
    compressed = coder.get("compressedData")
    root = WrapperNode(parent=None)

    if compressed or subset_count == 1:
        parse(root, codes)
    else:
        for ordinal in range(subset_count):
            node = WrapperNode(root, ordinal)
            parse(node, codes.copy())
            root.children.append(node)

    # Assign (delayed replication) factors

    global_factors = coder.get_delayed_replication_factors()

    def assign_factors(root: Node, global_factors: Dict[int, NDArray]) -> None:
        counter = Counter()

        def recurse(node, index):
            if isinstance(node, WrapperNode):
                for child in node.children:
                    recurse(child, index)
            elif isinstance(node, ReplicationNode):
                assert len(index) == node.level
                code = node.element.code
                at = counter[code]
                counter[code] += 1
                factor = global_factors[code][at]
                node.factors.insert(index, factor)
                for i in range(factor):
                    for child in node.children:
                        recurse(child, index + (i,))
            elif isinstance(node, LeafNode):
                pass
            elif isinstance(node, AssociationNode):
                assert node.level == 0
                if node.reuse_bitmap:
                    pass
                else:
                    if node.bitmap_size is None:
                        code = node.keys[0].element.code
                        assert code.FX == (0, 31)
                        at = counter[code]
                        counter[code] += 1
                        node.bitmap_size = global_factors[code][at]
                if node.value_size is None:
                    code = node.value_factor
                    assert code.FX == (0, 31)
                    at = counter[code]
                    counter[code] += 1
                    node.value_size = global_factors[code][at]
            else:
                assert False

        recurse(root, ())
        for code, count in counter.items():
            if remaining := len(global_factors[code]) - count:
                raise RuntimeError(
                    f"There are {remaining} unprocessed replication factors"
                )

    assign_factors(root, global_factors)

    # Make entries

    def make_entries(root: Node) -> Dict[str, DataEntry]:
        entries: Dict[str, DataEntry] = {}

        def recurse(node):
            if isinstance(node, (WrapperNode, ReplicationNode)):
                for child in node.children:
                    recurse(child)
            elif isinstance(node, (LeafNode, AssociationNode)):
                for key in node.keys:
                    try:
                        entry = entries[key.name]
                    except KeyError:
                        entry = DataEntry(key.name, flags=key.flags)
                        if entry.name in current_behaviour.assumed_scalar_elements:
                            entry.flags |= SCALAR
                        entry.uniform_element = key.element
                        entries[entry.name] = entry
                    else:
                        if entry.uniform_element:
                            if entry.uniform_element != key.element:
                                entry.uniform_element = None
            else:
                assert False

        recurse(root)
        return entries

    entries = make_entries(root)

    for code, array in global_factors.items():
        element = tables.elements[code]
        entry = entries[element.name]
        entry.shape = (array.size, 1)
        entry.array = np.reshape(array, entry.shape)

    # Assign bitmaps

    global_bitmap = coder.get_bitmap()

    def assign_bitmaps(root: Node, global_bitmap: NDArray) -> None:
        bitmap_for_reuse = None
        global_bitmap = copy(
            global_bitmap
        )  # parameter must be copied to be used as nonlocal
        for node in root.children:
            if not isinstance(node, AssociationNode):
                continue
            if node.reuse_bitmap:
                assert bitmap_for_reuse is not None
                node.bitmap = bitmap_for_reuse
            else:
                node.bitmap = global_bitmap[: node.bitmap_size]
                global_bitmap = global_bitmap[node.bitmap_size :]
                if node.for_reuse:
                    bitmap_for_reuse = node.bitmap
            assert node.bitmap is not None
            bitmap_count = numpy.count_nonzero(node.bitmap)
            assert node.value_size is not None
            if node.value_size > bitmap_count:
                # Although incorrect, assuming we can access bitmap_count
                # associated keys this shouldn't be a problem.
                pass
                # TODO: Issue a warning?
            elif node.value_size < bitmap_count:
                indices = numpy.arange(node.bitmap.size)
                cutoff = indices[node.bitmap][node.value_size]
                node.bitmap[cutoff:] = False
                # TODO: Maybe print a warning?

    assign_bitmaps(root, global_bitmap)

    # Resovle elements' (0-based) indices

    def resolve_indices(root: Node) -> Tuple[Dict[str, NDArray], int]:
        def extend(this, other):
            for k, v in other.items():
                this[k].extend(v)
            return this

        def recurse(node, counter, index):
            indices = defaultdict(list)
            if isinstance(node, WrapperNode):
                for child in node.children:
                    child_indices, counter = recurse(child, counter, index)
                    indices = extend(indices, child_indices)
            elif isinstance(node, ReplicationNode):
                assert node.level == len(index)
                for i in range(node.factors[index]):
                    for child in node.children:
                        child_indices, counter = recurse(child, counter, index + (i,))
                        indices = extend(indices, child_indices)
            elif isinstance(node, LeafNode):
                for k in node.keys:
                    indices[k.name].append(counter)
                    counter += 1
            elif isinstance(node, AssociationNode):
                node.bitmap_end = counter
            else:
                assert False
            return indices, counter

        indices, counter = recurse(root, 0, ())
        indices = {k: numpy.array(v) for k, v in indices.items()}
        return indices, counter

    if len(global_bitmap) > 0:
        indices, next_index = resolve_indices(root)

    # Make associations

    def make_associations(root: Node) -> Dict[Tuple[str, int], Association]:
        ranks = Counter()
        associations = {}
        for node in root.children:
            if not isinstance(node, AssociationNode):
                continue
            dtype: DTypeLike = numpy.dtype(float)
            if node.operator == 222000:
                e = node.element
                if e.scale == 0 and INT_UNITS.match(e.units):
                    dtype = numpy.dtype(int)
            ranks[node.element.name] += 1
            rank = ranks[node.element.name]
            assert node.bitmap is not None
            bitmap_offset = node.bitmap_end - len(node.bitmap)
            a = Association(
                node.element, rank, dtype, node.bitmap, bitmap_offset, indices
            )
            associations[node.element.name, rank] = a
        return associations

    associations = make_associations(root)

    # Do we still need this? TODO
    bitmap_nodes = [
        (i, c) for i, c in enumerate(root.children) if isinstance(c, AssociationNode)
    ]
    if bitmap_nodes:
        for i, node in reversed(bitmap_nodes):
            for element in (node.center, node.generating_application):
                if element.name not in indices:
                    indices[element.name] = numpy.array([next_index])
                else:
                    indices[element.name] = numpy.append(
                        indices[element.name], next_index
                    )
                next_index += 1

    # Append associated entries

    for a in associations.values():
        for pg in list(entries.values()):
            a_name = "->".join(repeat(a.element.name, a.element_rank))
            # if pg.flags & SCALAR:
            #     continue # TODO: are there cases where 'center' and 'originatingApplication' appear outside of bitmap sequence?
            if a.any_rank_set(pg.name):
                name = f"{pg.name}->{a_name}"
                ag = DataEntry(name, association=a, primary=pg)
                ag.flags = pg.flags
                if pg.uniform_element:
                    if a.element.code.FX != (0, 33):
                        ag.uniform_element = copy(a.element)
                        ag.uniform_element.units = pg.uniform_element.units
                    else:
                        ag.uniform_element = a.element
                entries[name] = ag
                a.entries[pg.name] = ag

    # Insert associated keys

    def insert_associated_keys(root: Node, associations) -> None:
        def recurse(node, associated):
            if isinstance(node, (WrapperNode, ReplicationNode)):
                for child in node.children:
                    recurse(child, associated)
            elif isinstance(node, LeafNode):
                name_suffix = "->".join(
                    repeat(associated.element.name, associated.element_rank)
                )
                copy_units = associated.element.code.FX != (0, 33)
                keys = []
                for key in node.keys:
                    keys.append(key)
                    if key.name in associated.entries:
                        if copy_units:
                            element = copy(associated.element)
                            assert key.element
                            element.units = key.element.units
                        else:
                            element = associated.element
                        key = Key(
                            f"{key.name}->{name_suffix}",
                            secondary=name_suffix,
                            element=element,
                        )
                        keys.append(key)
                node.keys = keys
            elif isinstance(node, AssociationNode):
                pass
            else:
                assert False

        for each in associations:
            recurse(root, each)

    insert_associated_keys(root, associations.values())

    # Resolve max. replication levels

    def resolve_max_levels(node: Node) -> None:
        if isinstance(node, (WrapperNode, ReplicationNode)):
            for child in node.children:
                resolve_max_levels(child)
                for name, child_max_level in child.max_levels.items():
                    max_level = node.max_levels.get(name, 0)
                    node.max_levels[name] = max(max_level, child_max_level)
        elif isinstance(node, (LeafNode, AssociationNode)):
            for key in node.keys:
                node.max_levels[key.name] = node.level
        else:
            assert False

    resolve_max_levels(root)

    # Return

    return root, entries, associations



#!/usr/bin/env python3

# Copyright 2005- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

# This example shows how to list all items in a BUFR message.
#
# Usage: ./items.py input.bufr

# flake8: noqa: F403
# flake8: noqa: F405

import sys

from eccodes import *
from eccodes.highlevel import *


def run_example(input, output=sys.stdout):
    reader = FileReader(input, CODES_PRODUCT_BUFR)
    for i, message in enumerate(reader, start=1):
        print("messageNumber", i, file=output)
        if message["compressedData"] or message["numberOfSubsets"] == 1:
            for k, v in message.items():
                print(k, v, file=output)
        else:
            for k, v in message.header.items():
                print(k, v, file=output)
            # For uncompressed multi-subset messages, print items one subset at a time. [1]
            for j, subset in enumerate(message.data, start=1):
                print("subsetNumber", j, file=output)
                for k, v in subset.items():
                    print(k, v, file=output)


if __name__ == "__main__":
    run_example(sys.argv[1])

# Footnotes:
#
# [1] This is not stricly necessary but, generally, it's easier to interpret
#     contents of ucompressed messages by looking at individual subsets, especially
#     so when subsets have different replication factors.



# Copyright 2022- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

# flake8: noqa: F405
#   ruff: noqa: F403

from .coder import INPUT_TEMPLATE_KEYS, TEMPLATE_KEYS, Coder
from .common import *
from .helpers import get_date, get_datetime, get_time, set_date, set_datetime, set_time
from .view import View


@dataclass
class Entry:
    name: str = ""
    value: Optional[ValueLike] = None
    dirty: bool = False
    computed: bool = False
    derive: Optional[Callable] = None
    project: Optional[Callable] = None


def derive_datetime(header, key):
    prefix = key.split("DateTime")[0]
    datetime = get_datetime(header, prefix=prefix)
    return datetime


def derive_date(header, key):
    prefix = key.split("Date")[0]
    date = get_date(header, prefix=prefix)
    return date


def derive_time(header, key):
    prefix = key.split("Time")[0]
    time = get_time(header, prefix=prefix)
    return time


def derive_rdbtime_datetime(header, key):
    date = derive_rdbtime_date(header, "rdbtimeDate")
    time = derive_time(header, "rdbtimeTime")
    datetime = date + time
    return datetime


def derive_rdbtime_date(header, key):
    assert key == "rdbtimeDate"
    string = header._coder.get(key, validate=False)
    y, m, d = string[0:4], string[4:6], string[6:8]
    date = np.datetime64(f"{y}-{m}-{d}", "D")
    return date


def project_datetime(header, key, value):
    prefix = key.split("DateTime")[0]
    set_datetime(header, value, prefix=prefix)


def project_date(header, key, value):
    prefix = key.split("Date")[0]
    set_date(header, value, prefix=prefix)


def project_time(header, key, value):
    prefix = key.split("Time")[0]
    set_time(header, value, prefix=prefix)


def project_edition_3_typical_year(header, key, value):
    assert key == "typicalYear"
    assert header["edition"] == 3
    header["typicalCentury"] = value // 100 + 1
    header["typicalYearOfCentury"] = value % 100


COMMON_COMPUTED_KEYS = set(
    [
        "typicalDateTime",
        "typicalDate",
        "typicalTime",
        "localDateTime",
        "localDate",
        "localTime",
        "rdbtimeDateTime",
        "rdbtimeDate",
        "rdbtimeTime",
        "rectimeTime",
        "md5Data",
    ]
)

_COMPUTED_ENTRIES = {
    None: [Entry(key, computed=True) for key in COMMON_COMPUTED_KEYS],
    3: [Entry("typicalYear", computed=True, project=project_edition_3_typical_year)],
    4: [
        Entry("typicalCentury", computed=True),
        Entry("typicalYearOfCentury", computed=True),
    ],
}

COMPUTED_ENTRIES = {}
for edition in _COMPUTED_ENTRIES.keys():
    COMPUTED_ENTRIES[edition] = {
        entry.name: entry for entry in _COMPUTED_ENTRIES[edition]
    }

for key, entry in COMPUTED_ENTRIES[None].items():
    if key == "rdbtimeDateTime":
        entry.derive = derive_rdbtime_datetime
    elif key == "rdbtimeDate":
        entry.derive = derive_rdbtime_date
    elif key.endswith("DateTime"):
        entry.derive = derive_datetime
        entry.project = project_datetime
    elif key.endswith("Date"):
        entry.derive = derive_date
        entry.project = project_date
    elif key.endswith("Time"):
        entry.derive = derive_time
        entry.project = project_time


class Header(View):
    """Provides access to BUFR header entries."""

    _coder: Coder
    _cache: Dict[str, Entry]

    def __init__(self, coder: Coder) -> None:
        self._coder = coder
        edition = cast(int, coder.get("edition", validate=False))
        self._cache = COMPUTED_ENTRIES[None].copy()
        self._cache.update(COMPUTED_ENTRIES[edition])

    def __contains__(self, key: str) -> bool:
        return (key in self._cache) or self._coder.is_defined(key, header_only=True)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Header):
            eq = False
            items1 = self.items(skip="read_only")
            items2 = other.items(skip="read_only")
            for (k1, v1), (k2, v2) in zip(items1, items2):
                if k1 != k2:
                    break
                if isinstance(v1, np.ndarray):
                    if not isinstance(v2, np.ndarray):
                        break
                    if not np.all(v1.data == v2.data):
                        break
                else:
                    if v1 != v2:
                        break
            else:
                eq = True
        else:
            eq = False
        return eq

    def __getitem__(self, key: str) -> ValueLike:
        try:
            entry = self._cache[key]
        except KeyError:
            value = self._coder.get(key, header_only=True)
            self._cache[key] = Entry(key, value)
        else:
            if entry.computed:
                assert entry.value is None
                if entry.derive:
                    value = entry.derive(self, key)
                else:
                    value = self._coder.get(key, validate=False)
            else:
                assert entry.value is not None
                value = entry.value
        return value

    def __setitem__(self, key: str, value: ValueLike) -> None:
        if key == "edition":
            current_edition = cast(int, self._coder.get("edition", validate=False))
            if value != current_edition:
                for k in COMPUTED_ENTRIES[current_edition].keys():
                    self._cache.pop(k, None)
                assert isinstance(value, int)
                self._cache.update(COMPUTED_ENTRIES[value])
        try:
            entry = self._cache[key]
        except KeyError:
            self._coder.set(key, value, header_only=True)
            self._cache[key] = Entry(key, value, dirty=True)
        else:
            if entry.computed:
                if entry.project:
                    entry.project(self, key, value)
                else:
                    self._coder.set(key, value, header_only=True)
            else:
                self._coder.set(key, value, header_only=True)
                entry.value = value
                entry.dirty = True

    def __str__(self) -> str:
        import pprint

        d = self.as_dict()
        return pprint.pformat(d, sort_dicts=False)

    def copy_to(self, other: "Header", skip_template_keys=False) -> None:
        """Copies common items from this header to `other`."""
        for key, value in self.items(skip="read_only"):
            if key == "unexpandedDescriptors":
                break
            try:
                other[key] = value
            except NotFoundError:
                continue
        if not skip_template_keys:
            if other._coder._baked_template:
                raise EncodingError("Cannot change baked template")
            self._coder.unpack()
            for skey, okey in zip(TEMPLATE_KEYS, INPUT_TEMPLATE_KEYS):
                if okey not in other._cache:
                    try:
                        value = self._coder.get(skey, validate=False)
                    except NotFoundError:
                        continue
                    else:
                        other._coder.set(okey, value)

    def get_count(self, key: str) -> int:
        """Returns the number of elements designated by the given key."""
        self.__getitem__(key)
        return 1

    def get_shape(self, key: str) -> Tuple[int, ...]:
        value = self.__getitem__(key)
        if isinstance(value, np.ndarray):
            shape = value.shape
        else:
            shape = ()
        return shape

    def get_size(self, key: str) -> int:
        value = self.__getitem__(key)
        if isinstance(value, np.ndarray):
            size = value.size
        else:
            size = 1
        return size

    def is_missing(self, key: str) -> bool:
        self.__getitem__(key)
        return False

    def items(self, ranked=False, **kwds) -> Iterator[Tuple[str, ValueLike]]:
        for key in self.keys(**kwds):
            yield key, self._coder.get(key, validate=False)

    def keys(self, ranked=False, **kwds) -> Iterator[str]:
        yield from self._coder.keys(header_only=True, **kwds)

    def set(self, key: str, value: ValueLike) -> None:
        self.__setitem__(key, value)

    def set_missing(self, key: str):
        self.__getitem__(key)
        raise ValueCannotBeMissingError(f"Header key '{key}' can't have missing value'")

    def as_dict(self, ranked=False, depth=0, **kwds) -> Dict:
        """Returns dict-like representation of the header items."""
        counts_only = kwds.pop("counts_only", False)
        shapes_only = kwds.pop("shapes_only", False)
        if counts_only:
            d = {k: self.get_count(k) for k in self.keys(ranked, **kwds)}
        elif shapes_only:
            d = {k: self.get_shape(k) for k in self.keys(ranked, **kwds)}
        else:
            d = dict(self.items(ranked, **kwds))
        return d

    def _commit(self) -> None:
        return  # [1]
        if self._cache:
            for key in [
                "edition",
                "bufrHeaderCentre",
                "masterTableNumber",
            ]:  # must be set first
                try:
                    value = self._cache.pop(key)
                except KeyError:
                    continue
                self._coder.set(key, value, validate=False)  # [2]
            for key, value in self._cache.items():
                self._coder.set(
                    key, value, validate=False, ignore_read_only_error=True
                )  # [2]

        # [1] Currently, when setting header keys, we pass the values right through
        #     into the coder, so there is no need to flush the cache before packing.
        #     But we will probably need this in the future when adding support for
        #     dynamically resizeable replication blocks, which will require full
        #     virtualisation of BUFR messages. TODO
        #
        # [2] Just a reminder that we don't need to use header_only=True here
        #     because the keys in the cache are header-only by definition.



# Copyright 2022- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

# flake8: noqa: F403
# flake8: noqa: F405
import sys

import pytest as pt

from eccodes.highlevel._bufr.message import *

if sys.platform.startswith("win32"):
    pt.skip("Not applicable on Windows", allow_module_level=True)


@pt.fixture(autouse=True)
def fixture(monkeypatch):
    import os

    this_dir = os.path.dirname(__file__)
    monkeypatch.chdir(this_dir)  # always run tests from the tests directory


def test_header_accessors():
    bufr = BUFRMessage(open("./sample-data/synop.bufr"))
    # Make sure we can access header section keys
    assert "localYear" in bufr.header
    bufr.header["localYear"] = 2025
    assert bufr.header["localYear"] == 2025
    # Make sure we can't access data section keys
    assert "year" not in bufr.header
    assert "year->code" not in bufr.header
    assert "year->percentConfidence" not in bufr.header
    assert "year->percentConfidence->code" not in bufr.header
    with pt.raises(KeyValueNotFoundError):
        bufr.header["year"]
    with pt.raises(KeyValueNotFoundError):
        bufr.header["year->code"]
    with pt.raises(KeyValueNotFoundError):
        bufr.header["year->percentConfidence"]
    with pt.raises(KeyValueNotFoundError):
        bufr.header["year->percentConfidence->code"]
    with pt.raises(KeyValueNotFoundError):
        bufr.header["year"] = 2025
    with pt.raises(KeyValueNotFoundError):
        bufr.header["year->percentConfidence"] = 100


def test_header_get_count():
    bufr = BUFRMessage("BUFR3_local")
    assert bufr.get_count("edition") == 1
    with pt.raises(KeyValueNotFoundError):
        bufr.header.get_count("year")
    with pt.raises(KeyValueNotFoundError):
        bufr.header.get_count("#1#year")


def test_header_is_missing():
    bufr = BUFRMessage("BUFR3_local")
    assert not bufr.is_missing("edition")
    with pt.raises(KeyValueNotFoundError):
        bufr.header.is_missing("year")


def test_header_set_missing():
    bufr = BUFRMessage("BUFR3_local")
    with pt.raises(ValueCannotBeMissingError):
        bufr.set_missing("edition")
    with pt.raises(KeyValueNotFoundError):
        bufr.header.set_missing("year")


def test_header_computed_keys():
    bufr = BUFRMessage("BUFR3")
    bufr["typicalDate"] = dt.date(2000, 1, 2)
    bufr["typicalTime"] = dt.time(3, 4)
    assert bufr["typicalDateTime"] == np.datetime64("2000-01-02T03:04")
    assert bufr["typicalDate"] == np.datetime64("2000-01-02")
    assert bufr["typicalTime"] == np.timedelta64(3 * 60 + 4, "m")
    bufr["typicalDate"] = dt.datetime(2000, 1, 2)
    bufr["typicalTime"] = dt.timedelta(hours=3, minutes=4)
    assert bufr["typicalDateTime"] == np.datetime64("2000-01-02T03:04")
    bufr["typicalDate"] = np.datetime64("2000-01-02", "D")
    bufr["typicalTime"] = np.timedelta64(3 * 60 + 4, "m")
    assert bufr["typicalDateTime"] == np.datetime64("2000-01-02T03:04")
    bufr["typicalDate"] = "2000-01-02"
    bufr["typicalTime"] = "03:04"
    assert bufr["typicalDateTime"] == np.datetime64("2000-01-02T03:04")
    with pt.raises(NotFoundError):
        bufr["localDateTime"]
    bufr = BUFRMessage("BUFR3_local")
    bufr["localYear"] = 2000
    bufr["localMonth"] = 1
    bufr["localDay"] = 2
    bufr["localHour"] = 3
    bufr["localMinute"] = 4
    assert bufr["localDateTime"] == np.datetime64("2000-01-02T03:04")
    assert bufr["localDate"] == np.datetime64("2000-01-02")
    assert bufr["localTime"] == np.timedelta64(3 * 60 + 4, "m")



# Copyright 2022- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import io
import warnings
from functools import cached_property

from .common import *
from .helpers import ensure_array, missing_of
from .tables import Tables, Version

# flake8: noqa: F405
#   ruff: noqa: F403


TEMPLATE_KEYS = dict.fromkeys(
    [
        "numberOfSubsets",
        "compressedData",
        "dataPresentIndicator",
        "delayedDescriptorAndDataRepetitionFactor",
        "delayedDescriptorReplicationFactor",
        "extendedDelayedDescriptorAndDataRepetitionFactor",
        "extendedDelayedDescriptorReplicationFactor",
        "shortDelayedDescriptorReplicationFactor",
        "unexpandedDescriptors",
    ]
)

INPUT_TEMPLATE_KEYS = dict.fromkeys(
    [  # [1]
        "numberOfSubsets",
        "compressedData",
        "inputDataPresentIndicator",
        "inputDelayedDescriptorAndDataRepetitionFactor",
        "inputDelayedDescriptorReplicationFactor",
        "inputExtendedDelayedDescriptorAndDataRepetitionFactor",
        "inputExtendedDelayedDescriptorReplicationFactor",
        "inputShortDelayedDescriptorReplicationFactor",
        "unexpandedDescriptors",
    ]
)

# [1] When creating a new message, template keys must be set in a specific
#     order. That's why we use a `dict` and not a `set`; `sets` don't
#     preserve insertion order.

ARRAY_KEYS = TEMPLATE_KEYS.copy()
ARRAY_KEYS.update(INPUT_TEMPLATE_KEYS)
ARRAY_KEYS.pop("numberOfSubsets")
ARRAY_KEYS.pop("compressedData")


class Coder:
    _handle: int = 0
    _unpacked: bool = False
    _subset_count: int = 0
    _baked_template: bool = False
    _compressed: bool = False
    _autorelease: bool = True
    _clone_handle: int = 0  # [1]
    _clone_retry_count: int = 0
    _last_extract_method: str = ""  # [2]

    # [1] We need an extra handle so that we can work around issues with repeated
    #     subset extractions (see ECC-2015 and ECC-2016).
    #
    # [2] We are keeping track of the latest extraction method because ecCodes
    #     currently doesn't handle cases with mixed extractions correctly; that
    #     is not without creating a new clone of the original handle (see ECC-2016).

    def __init__(self, source) -> None:
        if isinstance(source, io.IOBase):
            try:
                fileno = source.fileno()
            except OSError:
                fileno = None
            if fileno is None:
                raise TypeError(
                    "Expected file-like object with a file descriptor; got %s"
                    % type(source)
                )
            handle = codes_bufr_new_from_file(source)
            if not handle:
                raise EOFError()
        elif isinstance(source, bytes):
            handle = codes_new_from_message(source)
        elif isinstance(source, str):
            handle = codes_bufr_new_from_samples(source)
        elif isinstance(source, int):
            handle = source
        else:
            raise TypeError(
                "Expected file-like object, bytes array or a sample name; got %s"
                % type(source)
            )
        self._handle = handle
        self._autorelease = current_behaviour.autorelease_handle

    def __del__(self) -> None:
        if self._autorelease:
            self.release()

    def clone(self, subsets=None) -> "Coder":
        self.pack()
        if subsets is None:
            handle = codes_clone(self._handle)
            coder = Coder(handle)
        else:
            if not self._clone_handle:  # [1]
                self._clone_handle = codes_clone(self._handle)
                codes_set_long(self._clone_handle, "unpack", 1)
            try:  # [2]
                self._extract(subsets)
                handle = codes_clone(self._clone_handle)
            except (InternalError, ArrayTooSmallError) as error:  # [3]
                if self._clone_retry_count > 1:
                    raise error
                else:
                    codes_release(self._clone_handle)
                    self._clone_handle = 0
                    self._clone_retry_count += 1
                    self._last_extract_method = ""
                    coder = self.clone(subsets)
            else:
                self._clone_retry_count = 0
                coder = Coder(handle)
        return coder

    # [1] We need an extra handle so that we can work around issues with repeated
    #     subset extractions (see ECC-2015).
    #
    # [2] This is also related to ECC-2015. If the extraction of subsets fails,
    #     we try one more time with a new clone of the original handle. If that
    #     fails too, we raise an exception.
    #
    # [3] The ArrayTooSmallError is needed because of ECC-2025.

    def _extract(self, subsets):
        if isinstance(array := subsets, np.ndarray):
            if array.dtype == bool:
                subsets = np.arange(array.size)[array]
            elif array.dtype != int:
                message = "`subsets` must be an array of bool or int; got type %s"
                raise TypeError(message % array.dtype)
            start, stop = subsets[0], subsets[-1] + 1
            is_interval = subsets.size == stop - start
            is_interval = is_interval and np.all(subsets == np.arange(start, stop))
        elif isinstance(subsets, range):
            is_interval = subsets.step == 1
        elif isinstance(subsets, Sequence):
            subsets = list(subsets)
            is_interval = True
            for s, r in zip(subsets, range(subsets[0], subsets[0] + len(subsets))):
                if s != r:
                    is_interval = False
                    break
        elif isinstance(subsets, slice):
            start = 0 if subsets.start is None else subsets.start
            stop = self._subset_count if subsets.stop is None else subsets.stop  # [1]
            step = 1 if subsets.step is None else subsets.step
            subsets = range(start, stop, step)
            is_interval = subsets.step == 1
        else:
            message = "`subsets` must be an array, sequence, or a slice; got %s"
            raise TypeError(message % type(subsets))
        if is_interval:
            method = "SubsetInterval"
            start = subsets[0] + 1
            end = subsets[-1] + 1
            codes_set_long(self._clone_handle, "extractSubsetIntervalStart", start)
            codes_set_long(self._clone_handle, "extractSubsetIntervalEnd", end)
        else:
            method = "SubsetList"
            codes_set_long_array(
                self._clone_handle, "extractSubsetList", [s + 1 for s in subsets]
            )
        codes_set_long(self._clone_handle, "doExtractSubsets", 1)
        if (last := self._last_extract_method) and method != last:
            codes_release(self._clone_handle)
            self._clone_handle = codes_clone(self._handle)
            codes_set_long(self._clone_handle, "unpack", 1)
            self._last_extract_method = ""
            self._extract(subsets)
        self._last_extract_method = method

    # [1] Note that there is a bug in ecCodes where the extraction operation
    #     resets the original value of 'numberOfSubsets' to the number of
    #     extracted subsets. That's why it's important to use the subset count
    #     from the original handle, not the clone!

    def get_buffer(self) -> bytes:
        self.pack()
        bytes = codes_get_message(self._handle)
        return bytes

    def get_bitmap(self) -> NDArray:
        """Returns bitmap array as a boolean mask.

        Note that this is a concatenation of all bitmaps into a single, large array.
        """
        self.unpack()
        try:
            bitmap = codes_get_long_array(self._handle, "dataPresentIndicator")
        except NotFoundError:
            bitmap = np.array([])
        bitmap[:] = 1 - bitmap[:]  # [1]
        bitmap = np.ma.make_mask(bitmap, copy=False, shrink=False, dtype=bitmap.dtype)
        return bitmap

        # [1] We are flipping zeros and ones to follow the standard convention
        #     of 1 meaning True, and 0 False, not the other way arround.

    def get_delayed_replication_factors(self) -> Dict[int, NDArray]:
        """Returns a `dict` of concatenated delayed replication factors."""
        self.unpack()
        factors = {}
        for code, name in [
            (31000, "shortDelayedDescriptorReplicationFactor"),
            (31001, "delayedDescriptorReplicationFactor"),
            (31002, "extendedDelayedDescriptorReplicationFactor"),
            (31011, "delayedDescriptorAndDataRepetitionFactor"),
            (31012, "extendedDelayedDescriptorAndDataRepetitionFactor"),
        ]:
            try:
                array = codes_get_long_array(self._handle, name)
            except NotFoundError:
                continue
            factors[code] = array
        return factors

    def checkout(self, entry: DataEntry):
        if entry.flags & FACTOR:
            # In principle we shouldn't need this branch as we pre-load all delayed
            # replication factor arrays in build_tree(). However, some messages
            # can have wrongly encoded bitmap where the factor ends up having an
            # associated key (e.g., delayedReplicationFactor->percentConfidence).
            # TODO: Should we hide these keys from the user?
            array = codes_get_array(self._handle, entry.name)
            array = np.reshape(array, entry.shape)
        elif entry.flags & BITMAP:
            array = codes_get_array(self._handle, entry.name)
            entry.shape = (
                entry.shape[0],
                array.size,
            )  # infer the 2nd dimension earlier TODO
        elif entry.association:
            if entry.name.find("->associatedField") > 0:
                # Note that his works reliably only for ->associatedField keys.
                # For bitmap-associated keys, if some of the ranks don't have an associated
                # value, this will either fail (ECC-1272), or worse, return an incomplete
                # array (ECC-1689).
                array = codes_get_array(self._handle, entry.name)
                # array = self._ensure_correct_size(entry, array)
            else:
                assert entry.primary
                rank_mask = entry.association.rank_mask(entry.primary.name)
                if np.count_nonzero(rank_mask) == 0:
                    raise NotFoundError(entry.name)
                assert len(rank_mask) == entry.primary.shape[0]
                dtype = entry.association.element_dtype
                array = np.empty(entry.shape, dtype)
                fill_value = missing_of(dtype)
                for rank, is_set in enumerate(rank_mask, start=1):
                    if is_set:
                        # Note: bitmap-associated keys have to be retrieved one
                        # rank at a time (see ECC-1272).
                        if not self._compressed or entry.flags & SCALAR:
                            array[rank - 1] = codes_get(
                                self._handle, f"#{rank}#{entry.name}"
                            )
                        else:
                            array[rank - 1, :] = codes_get_array(
                                self._handle, f"#{rank}#{entry.name}"
                            )
                    else:
                        array[rank - 1] = fill_value
        elif entry.uniform_element and entry.uniform_element.code == 31021:  # [0]
            array = np.empty(entry.shape, int)
            if not self._compressed or entry.flags & SCALAR:
                for rank in range(1, entry.shape[0] + 1):
                    array[rank - 1] = codes_get_long(
                        self._handle, f"#{rank}#{entry.name}"
                    )
            else:
                for rank in range(1, entry.shape[0] + 1):
                    array[rank - 1, :] = codes_get_long_array(
                        self._handle, f"#{rank}#{entry.name}"
                    )
        else:
            ktype = None
            if entry.name == "second":
                if entry.uniform_element and entry.uniform_element.scale == 0:
                    ktype = int  # [1]
            array_or_list = codes_get_array(self._handle, entry.name, ktype)
            if isinstance(array_or_list, list) and isinstance(array_or_list[0], str):
                if not entry.uniform_element:
                    raise NotImplementedError(
                        "Non-uniform string keys are not supported yet: %s" % entry.name
                    )
                element = self._tables.elements[entry.uniform_element.code]
                byte_count, remainder = divmod(element.width, 8)
                assert remainder == 0
                dtype = np.dtype(("<U", byte_count))
            else:
                dtype = None
            array = ensure_array(array_or_list, dtype)
            if entry.name == "centre":  # [2]
                for rank in range(1, entry.shape[0] + 1):
                    array[rank - 1] = codes_get_long(self._handle, f"#{rank}#centre")
                array = array[0 : entry.shape[0]]
            array = self._ensure_correct_size(entry, array)
        return array

        # [0] Currently, ecCodes doesn't allow to retrieve ...->associatedFieldSignificance
        #     values (descriptor 31021) with a single call to codes_get_array(), so
        #     we have to get the values one rank at a time. See ECC-2098.
        #
        # [1] ecCodes infers native type of key 'second' (descriptor 004006) to
        #     be float, but this is not quite right. The default scale of this
        #     descriptor is 0, meaning it can't represent floating-point values.
        #     So, in the absence of 'Change scale' operator, the default native
        #     type should be int.
        #
        # [2] To work around ECC-1624, we have to get 'centre' values rank by rank.

    def commit(self, entry) -> None:
        key = entry.name
        array = entry.array
        array.data[array.mask] = array.fill_value
        if entry.association:
            a = entry.association
            rank_mask = a.rank_mask(entry.primary.name)
            array = array.reshape(entry.shape)
            for rank, value in enumerate(array, start=1):
                if rank_mask[rank - 1]:
                    if value.ndim == 1:
                        codes_set_array(self._handle, f"#{rank}#{key}", value.data)
                    else:
                        codes_set(self._handle, f"#{rank}#{key}", value)
        else:
            if array.dtype.type == np.str_:
                if array.size == 1:  # [1]
                    codes_set(self._handle, key, array.data[0][0])
                elif array.size > 1 and not np.any(array != array.data[0][0]):
                    codes_set_array(self._handle, key, array.data[0][0:1])
                else:
                    codes_set_array(self._handle, key, array.data.ravel())
            elif self._compressed:
                rank_count = entry.shape[0]
                for rank in range(1, rank_count + 1):
                    rank_array = array.data[rank - 1]
                    if np.any(rank_array != rank_array[0]):
                        codes_set_array(self._handle, f"#{rank}#{key}", rank_array)
                    else:
                        codes_set_array(
                            self._handle, f"#{rank}#{key}", rank_array[0:1]
                        )  # [2]
            else:
                if key == "centre":  # [3]
                    for rank in range(1, array.size + 1):
                        codes_set(self._handle, f"#{rank}#{key}", array.data[rank - 1])
                else:
                    try:
                        codes_set_array(self._handle, key, array.data)
                    except ArrayTooSmallError as error:  # [4]
                        if (
                            entry.uniform_element
                            and entry.uniform_element.code == 31021
                        ):
                            for rank in range(1, array.size + 1):
                                codes_set(
                                    self._handle, f"#{rank}#{key}", array.data[rank - 1]
                                )
                        else:
                            raise error

        if self._clone_handle:  # [5]
            codes_release(self._clone_handle)
            self._clone_handle = 0

        # [1] This is a workaround for ECC-1623.
        #
        # [2] If all values in an array are the same, encode only a single scalar value.
        #
        # [3] The key 'centre' is also an alias for the header key 'headerCentre',
        #     and calling codes_set_array() on an array of data elements would treat
        #     the first element as a header value, which is not what we want.
        #     To work around this we have to set the values rank by rank. See ECC-1624.
        #
        # [4] Currently, ecCodes doesn't allow to use codes_set_array() on
        #     ...->associatedFieldSignificance keys, so we have to set the
        #     values rank by rank. See ECC-2098.
        #
        # [5] If there had been any changes to the original handle since the
        #     last time we created a clone, make sure we re-create the clone
        #     from scratch next time around so that it picks the latest changes.
        #     Note that this is only needed as a workarond for ECC-2015 and ECC-2016.

    def _ensure_correct_size(self, entry, array):
        if entry.flags & SCALAR:
            assert entry.shape[1] == 1
            if array.size == entry.shape[0]:
                array = np.reshape(array, entry.shape)
            else:
                msg = f"'{entry.name}' is assumed to be a scalar element, but it has more than one value per rank"
                action = current_behaviour.on_assumed_scalar_element_invalid_size
                if action == "raise":
                    raise ValueError(msg)
                elif action == "warn":
                    warnings.warn(msg)
                elif action == "ignore":
                    pass
                else:
                    msg = f"'{action}' is not a valid action for `on_invalid_assumed_scalar_element_size`"
                    raise ValueError(msg)
                entry.flags = entry.flags & ~Flags.SCALAR
                entry.shape = (entry.shape[0], self._subset_count)
                array = self._ensure_correct_size(entry, array)
        elif self._compressed:
            # Workaround for ECC-428
            if array.size == entry.shape[0] and entry.shape[1] > 1:
                array = np.expand_dims(array, axis=1)
                array = np.broadcast_to(array, entry.shape).copy()  # [1]
            elif array.size > entry.shape[0] and array.size < entry.size:
                new_array = np.empty(entry.shape, array.dtype)
                offset = 0
                for rank in range(1, entry.shape[0] + 1):
                    size = codes_get_size(self._handle, f"#{rank}#{entry.name}")
                    assert size == entry.shape[1] or size == 1
                    new_array[rank - 1, :] = array[offset : offset + size]
                    offset += size
                array = new_array
            elif array.size == 1 and entry.size > 1:
                array = np.full(entry.shape, array[0])  # [2]
            else:
                array = np.reshape(array, entry.shape)
        return array

        # [1], [2] Avoid creating unnecessary copies in read-only mode. TODO

    @cached_property
    def _tables(self) -> Tables:
        return self.get_tables()

    def get(
        self, key: str, header_only=False, data_only=False, validate=True
    ) -> ValueLike:
        if validate:
            if header_only and data_only:
                raise ValueError("header_only and data_only can't be both True")
            if header_only:
                if not codes_bufr_key_is_header(self._handle, key):
                    raise NotFoundError(key)
                if key == "centre":
                    raise NotFoundError(
                        f"{key}: Did you mean bufrHeaderCentre? (see ECC-1624)"
                    )
                if "->" in key:  # [1]
                    raise NotFoundError(key)
            elif data_only:
                if codes_bufr_key_is_header(self._handle, key):
                    raise NotFoundError(key)
        try:
            if key in ARRAY_KEYS:
                value = codes_get_array(self._handle, key)
            else:
                try:
                    value = codes_get(self._handle, key)
                except ArrayTooSmallError:
                    value = codes_get_array(self._handle, key)
        except NotFoundError as error:
            error.msg = key
            raise error
        return value

        # [1] Make sure we don't accidentally leak data section keys/values via
        #     the Header class. This can happen if the user tries to get attribute
        #     of an associated key, e.g., 'pressure->percentConfidence->code'.
        #     This should be captured by codes_bufr_key_is_header(), but it's not!

    def get_tables(self) -> Tables:
        version = Version()
        version.master = cast(
            int, codes_get_long(self._handle, "masterTablesVersionNumber")
        )
        version.local = cast(
            int, codes_get_long(self._handle, "localTablesVersionNumber")
        )
        version.centre = cast(int, codes_get_long(self._handle, "bufrHeaderCentre"))
        version.subcentre = cast(
            int, codes_get_long(self._handle, "bufrHeaderSubCentre")
        )
        tables = Tables(version)
        return tables

    def is_defined(self, key: str, header_only=False, data_only=False) -> bool:
        assert not (header_only and data_only)
        if defined := codes_is_defined(self._handle, key):
            if header_only:
                defined = codes_bufr_key_is_header(self._handle, key)
            elif data_only:
                defined = not codes_bufr_key_is_header(self._handle, key)
        return defined

    def is_missing(self, entry: DataEntry, slice: slice) -> bool:
        assert slice.stop - slice.start > 0
        is_missing = 1
        for rank in range(slice.start + 1, slice.stop + 1):
            if not (
                is_missing := codes_is_missing(self._handle, f"#{rank}#{entry.name}")
            ):
                break
        return bool(is_missing)

    def set_missing(self, entry: DataEntry, slice):
        assert slice.stop - slice.start > 0
        for rank in range(slice.start + 1, slice.stop + 1):
            codes_set_missing(self._handle, f"#{rank}#{entry.name}")

    def keys(self, header_only=False, data_only=False, **kwargs) -> Iterator[str]:
        assert not (header_only and data_only)
        yield from keys_of(
            self._handle, header_only=header_only, data_only=data_only, **kwargs
        )

    def pack(self) -> bool:
        if self._unpacked:
            codes_set_long(self._handle, "pack", 1)
            total_length = codes_get_long(self._handle, "totalLength")
            try:
                codes_set_long(self._handle, "messageLength", min(total_length, 65535))
            except NotFoundError:
                pass
            return True
        else:
            return False

    def release(self) -> None:
        if self._handle:
            codes_release(self._handle)
            self._handle = 0
        if self._clone_handle:
            codes_release(self._clone_handle)
            self._clone_handle = 0

    def set(
        self,
        key: str,
        value: ValueLike,
        header_only=False,
        data_only=False,
        validate=True,
        ignore_read_only_error=False,
    ) -> None:
        if validate:
            if header_only and data_only:
                raise ValueError("header_only and data_only can't be both True")
            if self._baked_template and key in INPUT_TEMPLATE_KEYS:
                raise ReadOnlyError(key)
            if key in ("pack", "unpack"):
                message = (
                    "%sing via the '%s' key is disallowed. Call %s() method istead."
                )
                raise NotFoundError(message % (key.capitalize(), key, key))
            if header_only:
                if not codes_bufr_key_is_header(self._handle, key):
                    raise NotFoundError(key)
                if key == "centre":  # [1]
                    raise NotFoundError(
                        f"{key}: Did you mean bufrHeaderCentre? (see ECC-1624)"
                    )
            elif data_only:
                if codes_bufr_key_is_header(self._handle, key):
                    raise NotFoundError(key)
        if key == "unexpandedDescriptors":
            self._unpacked = True  # for messages created from samples
        try:
            if hasattr(value, "__iter__") and not isinstance(value, str):
                codes_set_array(self._handle, key, value)
            else:
                codes_set(self._handle, key, value)
        except ReadOnlyError as error:
            if ignore_read_only_error:
                pass
            else:
                error.msg += f": {key}"
                raise error
        if not header_only:
            if self._clone_handle:  # [2]
                codes_release(self._clone_handle)
                self._clone_handle = 0

        # [1] ecCodes allows use of 'centre' as an alias for 'bufrHeaderCentre', but
        #     we disallow it because it causes ambiguity when 'centre' simultaneously
        #     appears in the data section too, such as in the presence of bitmap operator
        #     for instance. It would be confusing if we returned combined values from
        #     both the header and the data section in a single array when accessing
        #     'centre' from the Message view. Unfortunately, right now this is exactly
        #     what ecCodes does (see ECC-1624).
        #
        # [2] If there had been any changes to the original handle's data section
        #     since the last time we had created the clone, make sure we re-create
        #     the clone from scratch next time around so that it picks up those changes.
        #     Note that this is only needed because of ECC-2015 and ECC-2016.

    def unpack(self) -> bool:
        if self._unpacked:
            return False
        else:
            codes_set_long(self._handle, "skipExtraKeyAttributes", 1)  # [1]
            codes_set_long(self._handle, "unpack", 1)
            self._unpacked = True
            return True

        # [1] We don't need to extract key attributes at the library level as we can
        #     infer all attributes, lazily, from the ElementTable.

    def write(self, file: BinaryIO) -> int:
        self.pack()
        codes_write(self._handle, file)
        size = codes_get_message_size(self._handle)
        return size


class KeysIterator(object):
    """A simple wrapper around `codes_keys_iterator()` function.

    This class is mostly for internal use. Generally speaking, users should
    rely on the higher-level interface provided by `Message`, `Header`
    and `Data` classes.
    """

    def __init__(self, msg_handle, bufr_only=True, skip: FlagsLike = None):
        skip_flags = ensure_flags(skip)
        if bufr_only:
            self._handle = codes_bufr_keys_iterator_new(msg_handle)
            self._next = codes_bufr_keys_iterator_next
            self._get_name = codes_bufr_keys_iterator_get_name
            self._delete = codes_bufr_keys_iterator_delete
        else:
            self._handle = codes_keys_iterator_new(msg_handle)
            self._next = codes_keys_iterator_next
            self._get_name = codes_keys_iterator_get_name
            self._delete = codes_keys_iterator_delete

        if skip_flags & Flags.CODED:
            codes_skip_coded(self._handle)
        if skip_flags & Flags.COMPUTED:
            codes_skip_computed(self._handle)
        if skip_flags & Flags.READ_ONLY:
            codes_skip_read_only(self._handle)

    def __del__(self):
        self._delete(self._handle)

    def __iter__(self):
        return self

    def __next__(self):
        if self._next(self._handle):
            return self._get_name(self._handle)
        else:
            raise StopIteration


def keys_of(
    msg_handle,
    bufr_only=True,
    header_only=False,
    data_only=False,
    skip: FlagsLike = None,
) -> Iterator[str]:
    """Returns an iterator over the keys of the message.

    By default, the iteration runs over all keys, which includes keys from both
    the header and the data section. Optionally, user can constrain the
    iterator to only return keys from one of the sections by setting
    `header_only` or `data_only` keyword argument.

    Note that in order to access data section keys, the message must be
    unpacked. It's user's responsibility to do so before calling this function.
    If the message hasn't been unpacked, trying to access data keys will raise
    an exception.

    When `bufr_only` is True, the iterator returns BUFR-specific keys only; if
    False, all generic keys are returned instead.

    This function is mostly for internal use. Generally speaking, users should
    rely on the higher-level interface provided by `Message`, `Header`
    and `Data` classes.
    """
    if header_only and data_only:
        raise ValueError(
            "Keyword arguments `header_only` and `data_only` can't both be True"
        )

    data_keys = False
    keys = KeysIterator(msg_handle, bufr_only, skip)
    data_keys_unaccessible = (
        "Cannot access data keys because message hasn't been unpacked. "
    )
    data_keys_unaccessible += "If you want to access header keys only, set keyword argument `header_only` to True."

    if bufr_only:
        for key in keys:
            if key == "unexpandedDescriptors":
                next_key = next(keys, None)
                if next_key is None:
                    if not header_only:
                        raise RuntimeError(data_keys_unaccessible)
                else:
                    data_keys = True
                if header_only:
                    yield key
                    break
                elif data_only:
                    key = next_key
                else:
                    yield key
                    key = next_key
            if data_only and not data_keys:
                continue
            yield key
    else:
        for key in keys:
            if key == "dataKeys":
                next_key = next(keys, None)
                if next_key == "section4Padding":
                    if not header_only:
                        raise RuntimeError(data_keys_unaccessible)
                else:
                    data_keys = True
                if header_only:
                    yield key
                    key = next_key
                elif data_only:
                    key = next_key
                else:
                    yield key
                    key = next_key
            elif key == "section4Padding":
                data_keys = False
            if (header_only and data_keys) or (data_only and not data_keys):
                continue
            yield key



"""
A High-level interface for en/decoding of BUFR files.
"""

from .common import change_behaviour, get_behaviour, set_behaviour
from .helpers import missing_of
from .message import BUFRMessage

__all__ = [
    "change_behaviour",
    "get_behaviour",
    "set_behaviour",
    "missing_of",
    "BUFRMessage",
]



# Copyright 2022- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import csv
import ctypes
import os
import re
import sys
from collections import ChainMap, UserDict
from dataclasses import dataclass
from pathlib import Path
from typing import Iterator, List, Tuple, Union

import eccodes
import gribapi


class Code(int):
    def __new__(cls, code: int):
        return int.__new__(cls, code)

    def __repr__(self) -> str:
        return "'%s'" % str(int(self)).zfill(6)

    @property
    def F(self) -> int:
        return int(self) // 100000

    @property
    def X(self) -> int:
        return int(self) // 1000 % 100

    @property
    def Y(self) -> int:
        return int(self) % 1000

    @property
    def FX(self) -> Tuple[int, int]:
        return self.F, self.X


CodeLike = Union[Code, int]


@dataclass
class Descriptor:
    code: Code = 0
    name: str = ""


@dataclass
class Element(Descriptor):
    units: str = ""
    scale: int = 0
    reference: int = 0
    width: int = 0


@dataclass(unsafe_hash=True)
class Version:
    master: int = 0
    local: int = 0
    centre: int = 0
    subcentre: int = 0


class _Singleton(type(UserDict)):
    cache = {}

    def __call__(cls, *args):
        try:
            obj = cls.cache[cls, args]
        except KeyError:
            obj = super().__call__(*args)
            cls.cache[cls, args] = obj
        return obj


class Table(UserDict, metaclass=_Singleton):
    def __init__(self, version: Version, local: bool) -> None:
        base_path = Path("bufr", "tables", "0")
        if local:
            dashed_version_number = "%d-%d" % (version.master, version.local)
            path = base_path.joinpath(
                "local",
                dashed_version_number,
                str(version.centre),
                str(version.subcentre),
            )
            if not gribapi.grib_full_defs_path(str(path / "element.table")):
                path = base_path.joinpath(
                    "local",
                    str(version.local),
                    str(version.centre),
                    str(version.subcentre),
                )
        else:
            path = base_path.joinpath("wmo", str(version.master))
        self.version = version
        self.path = path
        self.data = {}


class ElementTable(Table):
    def __init__(self, version: Version, local: bool) -> None:
        super().__init__(version, local)
        self.path /= "element.table"
        try:
            text = codes_read_file(self.path)
        except RuntimeError as e:
            if local:
                return  # ignore non-existing local tables
            else:
                raise e
        reader = csv.reader(text.splitlines(), delimiter="|")
        next(reader)  # skip header row
        for r in reader:
            c = Code(int(r[0]))
            e = Element(c, r[1], r[4], int(r[5]), int(r[6]), int(r[7]))
            self.data[c] = e

    def __getitem__(self, code: CodeLike) -> Element:
        try:
            return self.data[code]
        except KeyError:
            message = "No such element in tables version %s (%s): %s"
            raise KeyError(message % (self.version, self.path, code))


class SequenceTable(Table):
    def __init__(self, version: Version, local: bool) -> None:
        super().__init__(version, local)
        self.path /= "sequence.def"
        try:
            text = codes_read_file(self.path)
        except RuntimeError as e:
            if local:
                return  # ignore non-existing local tables
            else:
                raise e
        tokens = re.split("]\n?", text)
        tokens = tokens[:-1]  # strip the last, empty token
        for token in tokens:
            pair = token.split("= [")
            code = int(pair[0].strip().strip('"'))
            sequence = [Code(c) for c in pair[1].split(",")]
            self.data[code] = sequence

    def __getitem__(self, code: CodeLike) -> List[Code]:
        try:
            return self.data[code]
        except KeyError:
            message = "No such element in tables version %s (%s): %s"
            raise KeyError(message % (self.version, self.path, code))


class CombinedTable(Table):
    def __init__(self, table_class, version):
        self.master = table_class(version, False)
        self.local = table_class(version, True)
        self.version = version
        self.data = ChainMap(self.local.data, self.master.data)  # [1]

    # [1] Note that the individual mappings in the ChainMap are searched from first to last.


class Tables:
    def __init__(self, version):
        self.sequences = CombinedTable(SequenceTable, version)
        self.elements = CombinedTable(ElementTable, version)
        self.version = version

    def expand_codes(
        self, code_s: Union[CodeLike, List[CodeLike]], recursive=True
    ) -> Iterator[Code]:
        if isinstance(code_s, int):
            codes = [code_s]
        else:
            codes = code_s
        for c in map(Code, codes):
            if c.F == 3:
                sequence_codes = self.sequences[c]
                if recursive:
                    yield from self.expand_codes(sequence_codes)
                else:
                    yield from sequence_codes
            else:
                yield c

    def expand_descriptors(
        self, codes_s: Union[CodeLike, List[CodeLike]], recursive=True
    ) -> Iterator[Descriptor]:
        previous_code = Code(0)
        for code in self.expand_codes(codes_s, recursive):
            name = None
            if code.F == 0:
                try:
                    name = self.elements[code].name
                except KeyError as error:
                    if previous_code.FX == (2, 6):  # Operator: Signify data width
                        # TODO: raise warning
                        name = ""
                    else:
                        raise error
                previous_code = code
            yield Descriptor(code, name)


libcname = "msvcrt" if sys.platform.startswith("win") else None

libc = ctypes.CDLL(
    libcname, winmode=0
)  # automatically finds and loads the C standard library

fseek = libc.fseek
fseek.argtypes = [ctypes.c_void_p, ctypes.c_long, ctypes.c_int]
fseek.restype = ctypes.c_int
SEEK_SET, SEEK_END = 0, 2

ftell = libc.ftell
ftell.argtypes = [ctypes.c_void_p]
ftell.restype = ctypes.c_long

fread = libc.fread
fread.argtypes = [ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t, ctypes.c_void_p]
fread.restype = ctypes.c_size_t

fclose = libc.fclose
fclose.argtypes = [ctypes.c_void_p]
fclose.restype = ctypes.c_int

libeccodes = ctypes.CDLL(eccodes.codes_get_library_path())

codes_fopen = libeccodes.codes_fopen
codes_fopen.restype = ctypes.c_void_p


def codes_has_file(path: Union[str, os.PathLike]) -> bool:
    stream = libeccodes.codes_fopen(str(path).encode(), b"r")
    if stream:
        libc.fclose(stream)
        has = True
    else:
        has = False
    return has


def codes_read_file(path: Union[str, os.PathLike]) -> str:
    full_path = gribapi.grib_full_defs_path(str(path))
    stream = libeccodes.codes_fopen(full_path.encode(), b"r")
    if not stream:
        raise RuntimeError(f"Tables path does not exist: {path}")
    try:
        fseek(stream, 0, SEEK_END)
        buffer_size = ftell(stream)
        buffer = ctypes.create_string_buffer(buffer_size)
        fseek(stream, 0, SEEK_SET)
        read_size = fread(buffer, ctypes.sizeof(ctypes.c_char), buffer_size, stream)
        string = buffer.raw[:read_size].decode()
    finally:
        fclose(stream)
    return string



#!/usr/bin/env python3

# Copyright 2022- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

# This example shows how to read key attributes.
#
# Usage: attributes.py input.bufr

# flake8: noqa: F403
# flake8: noqa: F405

import sys

from eccodes import *
from eccodes.highlevel import *

ATTRIBUTES = ["code", "units", "scale", "reference", "width"]  # See BUFR Code Table B


def run_example(input, output=sys.stdout):
    # Loop over the messages in the file
    for number, bufr in enumerate(FileReader(input, CODES_PRODUCT_BUFR), start=1):
        print(f"messageNumber: {number}", file=output)

        # Loop over all keys in the message
        for name in bufr.keys():
            # Print element's attributes. Attributes themselves are keys.
            for attribute in ATTRIBUTES:
                try:
                    key = f"{name}->{attribute}"
                    value = bufr[key]
                    print(f"{key}: {value}", file=output)
                except eccodes.KeyValueNotFoundError:
                    pass


if __name__ == "__main__":
    run_example(sys.argv[1])



# Copyright 2022- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime as dt

from .coder import Coder
from .common import *
from .data import Data
from .header import Header
from .view import View

# flake8: noqa: F405
#   ruff: noqa: F403


class Message(View):
    def __init__(self, source) -> None:
        if isinstance(source, Coder):
            coder = source
        else:
            coder = Coder(source)
        self.header = Header(coder)
        self.data = Data(coder)
        self._coder = coder

    def __contains__(self, key: str) -> bool:
        """Return True if `key` is defined, otherwise return False."""
        return (key in self.header) or (key in self.data)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self._coder.release()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, BUFRMessage):
            eq = (self.header == other.header) and (self.data == other.data)
        else:
            eq = False
        return eq

    def __getitem__(self, key: str) -> ValueLike:
        if self._coder._unpacked:
            try:
                value = cast(ValueLike, self.data[key])
            except NotFoundError:
                value = self.header[key]
        else:
            try:
                value = self.header[key]
            except NotFoundError:
                value = cast(ValueLike, self.data[key])
        return value

    def __setitem__(self, key: str, value: ValueLike) -> None:
        if self._coder._unpacked:
            try:
                self.data[key] = value
            except NotFoundError:
                self.header[key] = value
        else:
            try:
                self.header[key] = value
            except NotFoundError:
                self.data[key] = value

    def __str__(self) -> str:
        import pprint

        d = self.as_dict()
        return pprint.pformat(d, sort_dicts=False)

    @classmethod
    def from_samples(cls, sample_name):
        """Creates a new message from the sample."""
        return cls(Coder(sample_name))

    def copy(
        self, *, subsets: Optional[Union[NDArray, Sequence, slice]] = None
    ) -> "BUFRMessage":
        """Returns a new copy of the message.

        If the optional argument `subsets` is specified, only the selected subsets
        will be copied. `subsets` can be a mask, an array or a sequence of (0-based)
        subset indices, or a slice. For conveniece, `subsets` can also be a slice
        with numpy's `datetime64` objects, or with the built-in `datetime` objects.
        For such slices the stop bounds are inclusive.
        """
        self._commit()  # [1]
        if (
            subsets is not None
            and isinstance(subsets, slice)
            and (
                isinstance(subsets.start, (np.datetime64, dt.datetime))
                or isinstance(subsets.stop, (np.datetime64, dt.datetime))
            )
        ):
            datetime = self.data.get_datetime(rank=1)
            start = dt.datetime.min if subsets.start is None else subsets.start
            stop = dt.datetime.max if subsets.stop is None else subsets.stop
            mask = np.logical_and(datetime >= start, datetime <= stop)
            coder = self._coder.clone(mask)
        else:
            coder = self._coder.clone(subsets)
        message = BUFRMessage(coder)
        return message

        # [1] In theory we could avoid committing (and hence packing) if we
        #     copied the "original" data plus Data's cache and tree structure. TODO

    def copy_to(
        self,
        other: "BUFRMessage",
        header_only=False,
        data_only=False,
        skip_template_keys=False,
    ) -> None:
        """Copies common keys/values from this message to `other`."""
        if header_only and data_only:
            raise ValueError("`header_only` and `data_only` can't both be true")
        if not data_only:
            self.header.copy_to(other.header, skip_template_keys)
        if not header_only:
            self.data.copy_to(other.data)

    def get_count(self, key: str) -> int:
        """Returns the number of elements designated by the given key."""
        if self._coder._unpacked:
            try:
                count = self.data.get_count(key)
            except NotFoundError:
                count = self.header.get_count(key)
        else:
            try:
                count = self.header.get_count(key)
            except NotFoundError:
                count = self.data.get_count(key)
        return count

    def get_shape(self, key: str) -> Tuple[int, ...]:
        """Returns shape of the given key."""
        if self._coder._unpacked:
            try:
                shape = self.data.get_shape(key)
            except NotFoundError:
                shape = self.header.get_shape(key)
        else:
            try:
                shape = self.header.get_shape(key)
            except NotFoundError:
                shape = self.data.get_shape(key)
        return shape

    def get_size(self, key: str) -> int:
        """Returns size of the given key."""
        if self._coder._unpacked:
            try:
                size = self.data.get_size(key)
            except NotFoundError:
                size = self.header.get_size(key)
        else:
            try:
                size = self.header.get_size(key)
            except NotFoundError:
                size = self.data.get_size(key)
        return size

    def is_missing(self, key: str) -> bool:
        """Checks the key is set to missing value.

        If the value is an array, returns true only if all elements in the array
        are set to missing value.
        """
        if self._coder._unpacked:
            try:
                is_missing = self.data.is_missing(key)
            except NotFoundError:
                is_missing = self.header.is_missing(key)
        else:
            try:
                is_missing = self.header.is_missing(key)
            except NotFoundError:
                is_missing = self.data.is_missing(key)
        return is_missing

    def set(self, key: str, value: ValueLike) -> None:
        if self._coder._unpacked:
            try:
                self.data.set(key, value)
            except NotFoundError:
                self.header.set(key, value)
        else:
            try:
                self.header.set(key, value)
            except NotFoundError:
                self.data.set(key, value)

    def set_missing(self, key: str) -> None:
        """Sets the key to missing value.

        If the value is an array, sets all elements to missing value.
        """
        if self._coder._unpacked:
            try:
                self.data.set_missing(key)
            except NotFoundError:
                self.header.set_missing(key)
        else:
            try:
                self.header.set_missing(key)
            except NotFoundError:
                self.data.set_missing(key)

    def get_buffer(self) -> bytes:
        """Returns a buffer containing the encoded message."""
        self._commit()
        return self._coder.get_buffer()

    def as_dict(self, ranked=False, depth=0, **kwds) -> Dict:
        """Returns dict-like representation of the message items."""
        header_only = kwds.pop("header_only", False)
        data_only = kwds.pop("data_only", False)
        if not data_only:
            h = self.header.as_dict(ranked, depth - 1, **kwds)
        if not header_only:
            d = self.data.as_dict(ranked, depth - 1, **kwds)
        if header_only:
            m = h
        elif data_only:
            m = d
        else:
            if depth > 0:
                m = {"header": h, "data": d}
            else:
                m = h | d
        return m

    def items(self, ranked=False, **kwds) -> Iterator[Tuple[str, ValueLike]]:
        yield from self.header.items(ranked, **kwds)
        yield from self.data.items(ranked, **kwds)

    def keys(self, ranked=False, **kwds) -> Iterator[str]:
        yield from self.header.keys(ranked, **kwds)
        yield from self.data.keys(ranked, **kwds)

    def pack(self) -> None:
        """Encodes the contents of the data section into an internal buffer.

        Note that this happens automatically whenever needed (before writing
        to a file, before copying, etc.), so users don't need to call this method
        explicitly.
        """
        self._commit()
        self._coder.pack()

    def unpack(self) -> None:
        """Decodes the contents of the data section into an internal representation.

        Note that this happens automatically whenever needed (when accessing data
        section keys for the first time, before copying, etc.), so users don't need
        to call this method explicitly.
        """
        self._coder.unpack()

    def write(self, file: BinaryIO) -> int:
        import warnings

        warnings.warn(
            "Method write() is deprecated. Use method write_to() instead.",
            DeprecationWarning,
        )
        return self.write_to(file)

    def write_to(self, file: BinaryIO) -> int:
        """Encodes the message (if not already encoded) and writes it to a file.

        Returns the size of the encoded message in bytes.

        `file` must be a file-like object.
        """
        self._commit()
        size = self._coder.write(file)
        return size

    def _commit(self) -> None:
        if current_behaviour.update_header_from_data_before_packing:
            self.update_header_from_data(skip_dirty=True)
        self.header._commit()
        self.data._commit()

    @property
    def _handle(self) -> int:
        return self._coder._handle

    def update_header_from_data(self, skip_dirty=False) -> None:
        """Updates some of the header keys based on keys from the data section.

        This method sets header keys such as 'typicalYear', 'localMonth',
        'localLatitude', etc., based on keys from the data section: 'year',
        'month', 'latitude', etc.

        If there are no datetime-related keys in message's data section, or if
        some of the keys that are needed to construct a complete datetime object
        are missing, the header keys are not updated.

        In the case of multi-subset messages we use the min value across all subsets.

        Note that this method is called automatically before packing, so users typically
        don't need to call this method explicitly. However, if this behaviour is not
        desirable, it can be switched off by setting the `Behaviour` option
        `update_header_from_data_before_packing` to `False`.
        """
        datetime_counts = {}
        for key in ("year", "month", "day", "hour", "minute"):
            try:
                entry = self.data._entries[key]
            except KeyError:
                has_datetime = False
                break
            if entry.array is None:
                has_datetime = False
                break
            else:
                datetime_counts[key] = self.data.get_count(key)
        else:
            if all(c == datetime_counts["year"] for c in datetime_counts.values()):
                datetime_array = self.data.get_datetime()
            else:
                max_common_rank = min(datetime_counts.values())
                datetime_array = self.data.get_datetime(
                    rank=slice(1, max_common_rank + 1)
                )
            min_datetime = np.min(datetime_array)
            if min_datetime is np.ma.masked:
                has_datetime = False
            else:
                datetime = min_datetime.item()
                has_datetime = True

        if has_datetime:
            items = ["Year", "YearOfCentury", "Day", "Hour", "Minute", "Second"]
            if not (_any_dirty(self.header, "typical", items) and skip_dirty):
                if self.header["edition"] == 3:
                    self.header["typicalYearOfCentury"] = datetime.year % 100
                else:
                    self.header["typicalYear"] = datetime.year
                self.header["typicalMonth"] = datetime.month
                self.header["typicalDay"] = datetime.day
                self.header["typicalHour"] = datetime.hour
                self.header["typicalMinute"] = datetime.minute
                self.header["typicalSecond"] = datetime.second

        if self.header["section2Present"] and self.header["bufrHeaderCentre"] == 98:
            if has_datetime:
                items = ["Year", "Day", "Hour", "Minute", "Second"]
                if not (skip_dirty and _any_dirty(self.header, "local", items)):
                    self.header["localYear"] = datetime.year
                    self.header["localMonth"] = datetime.month
                    self.header["localDay"] = datetime.day
                    self.header["localHour"] = datetime.hour
                    self.header["localMinute"] = datetime.minute
                    self.header["localSecond"] = datetime.second
            now = dt.datetime.now(dt.timezone.utc)
            for prefix in ("rdbtime", "rectime"):
                items = ["Day", "Hour", "Minute", "Second"]
                if not (skip_dirty and _any_dirty(self.header, prefix, items)):
                    self.header[f"{prefix}Day"] = now.day
                    self.header[f"{prefix}Hour"] = now.hour
                    self.header[f"{prefix}Minute"] = now.minute
                    self.header[f"{prefix}Second"] = now.second
            try:
                lat_entry = self.data._entries["latitude"]
                lon_entry = self.data._entries["longitude"]
            except KeyError:
                has_latlon = False
            else:
                if has_latlon := (lat_entry.array is not None) and (
                    lon_entry.array is not None
                ):
                    lat = lat_entry.array.ravel()
                    lon = lon_entry.array.ravel()
            if self.header["compressedData"]:
                if cast(int, self.header["numberOfSubsets"]) > 1:
                    if has_latlon:
                        items = ["Latitude1", "Latitude2", "Longitude1", "Longitude2"]
                        if not (skip_dirty and _any_dirty(self.header, "local", items)):
                            self.header["localLatitude1"] = numpy.min(lat)
                            self.header["localLatitude2"] = numpy.max(lat)
                            self.header["localLongitude1"] = numpy.min(lon)
                            self.header["localLongitude2"] = numpy.max(lon)
                    self.header["localNumberOfObservations"] = self.header[
                        "numberOfSubsets"
                    ]
                else:
                    if has_latlon:
                        items = ["Latitude1", "Longitude1"]
                        if not (skip_dirty and _any_dirty(self.header, "local", items)):
                            self.header["localLatitude1"] = lat[0]
                            self.header["localLongitude1"] = lon[0]
            else:
                if has_latlon:
                    items = ["Latitude1", "Longitude1"]
                    if not (skip_dirty and _any_dirty(self.header, "local", items)):
                        self.header["localLatitude"] = lat[0]
                        self.header["localLongitude"] = lon[0]


def _any_dirty(header: Header, prefix: str, keys: Sequence[str]) -> bool:
    for key in keys:
        try:
            entry = header._cache[f"{prefix}{key}"]
        except KeyError:
            continue
        else:
            if entry.dirty:
                return True
    else:
        return False


BUFRMessage = Message



# Copyright 2022- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

# flake8: noqa: F403
# flake8: noqa: F405
import sys
import warnings

import numpy as np
import pytest as pt

from eccodes import *
from eccodes.highlevel._bufr.message import *

if sys.platform.startswith("win32"):
    pt.skip("Not applicable on Windows", allow_module_level=True)


@pt.fixture(autouse=True)
def fixture(monkeypatch):
    import os

    this_dir = os.path.dirname(__file__)
    monkeypatch.chdir(this_dir)  # always run tests from the tests directory


def test_data_key():
    assert Key.from_string("year") == Key(name="year")
    assert Key.from_string("year->code") == Key(name="year", attribute="code")
    assert Key.from_string("year->units") == Key(name="year", attribute="units")
    assert Key.from_string("year->reference") == Key(name="year", attribute="reference")
    assert Key.from_string("year->scale") == Key(name="year", attribute="scale")
    assert Key.from_string("year->width") == Key(name="year", attribute="width")
    assert Key.from_string("year->percentConfidence") == Key(
        name="year->percentConfidence", primary="year", secondary="percentConfidence"
    )
    assert Key.from_string("year->percentConfidence->code") == Key(
        name="year->percentConfidence",
        primary="year",
        secondary="percentConfidence",
        attribute="code",
    )
    assert Key.from_string("#1#year") == Key(name="year", rank=1)
    assert Key.from_string("#1#year->code") == Key(
        name="year", rank=1, attribute="code"
    )
    assert Key.from_string("#1#year->percentConfidence") == Key(
        name="year->percentConfidence",
        rank=1,
        primary="year",
        secondary="percentConfidence",
    )
    assert Key.from_string("#1#year->percentConfidence->code") == Key(
        name="year->percentConfidence",
        rank=1,
        primary="year",
        secondary="percentConfidence",
        attribute="code",
    )
    assert "year" == Key(name="year").string
    assert "year->code" == Key(name="year", attribute="code").string
    assert "year->units" == Key(name="year", attribute="units").string
    assert "year->reference" == Key(name="year", attribute="reference").string
    assert "year->scale" == Key(name="year", attribute="scale").string
    assert "year->width" == Key(name="year", attribute="width").string
    assert (
        "year->percentConfidence"
        == Key(
            name="year->percentConfidence",
            primary="year",
            secondary="percentConfidence",
        ).string
    )
    assert (
        "year->percentConfidence->code"
        == Key(
            name="year->percentConfidence",
            primary="year",
            secondary="percentConfidence",
            attribute="code",
        ).string
    )
    assert "#1#year" == Key(name="year", rank=1).string
    assert "#1#year->code" == Key(name="year", rank=1, attribute="code").string
    assert (
        "#1#year->percentConfidence"
        == Key(
            name="year->percentConfidence",
            rank=1,
            primary="year",
            secondary="percentConfidence",
        ).string
    )
    assert (
        "#1#year->percentConfidence->code"
        == Key(
            name="year->percentConfidence",
            rank=1,
            primary="year",
            secondary="percentConfidence",
            attribute="code",
        ).string
    )
    with pt.raises(NotFoundError):
        Key.from_string("/pressure=85000/airTemperature")


def test_data_accessors():
    bufr = BUFRMessage(open("./sample-data/synop.bufr"))
    # Header keys (should not be accessible from the Data class)
    assert "edition" not in bufr.data
    with pt.raises(NotFoundError):
        bufr.data["edition"]
    with pt.raises(NotFoundError):
        bufr.data["edition"] = 3
    # Data keys: defined
    assert "year" in bufr.data
    assert "year->code" in bufr.data
    assert "year->percentConfidence" in bufr.data
    assert "year->percentConfidence->code" in bufr.data
    assert "#1#year" in bufr.data
    assert "#1#year->code" in bufr.data
    assert "#1#year->percentConfidence" in bufr.data
    assert "#1#year->percentConfidence->code" in bufr.data
    assert bufr.data["year"] == 2009
    assert bufr.data["year->code"] == 4001
    assert bufr.data["year->percentConfidence"] == 70
    assert np.all(
        bufr.data["cloudAmount"] == np.ma.masked_values([8, -1, -1, -1, -1], -1)
    )
    # assert bufr.data['year->percentConfidence->code'] == TODO
    assert bufr.data["#1#year"] == 2009
    assert bufr.data["#1#year->code"] == 4001
    assert bufr.data["#1#year->percentConfidence"] == 70
    assert bufr.data["#1#cloudAmount"] == 8
    # assert bufr.data['#1#year->percentConfidence->code'] == TODO
    bufr.data["year"] = 2025
    bufr.data["year->percentConfidence"] = 100
    bufr.data["cloudAmount"] = [1, 2, 3, 4, 5]
    bufr.data["#1#year"] = 2025
    bufr.data["#1#year->percentConfidence"] = 100
    bufr.data["#1#cloudAmount"] = 8
    with pt.raises(ReadOnlyError):
        bufr.data["year->code"] = 0
    with pt.raises(ReadOnlyError):
        bufr.data["#1#year->code"] = 0
    # with pt.raises(ReadOnlyError): bufr.data['#1#year->percentConfidence->code'] = TODO
    # Data keys: undefined
    assert "year->code->code" not in bufr.data
    assert "year->percentConfidence->code->code" not in bufr.data
    assert "#2#year" not in bufr.data
    assert "#2#year->percentConfidence" not in bufr.data
    assert "#2#year->percentConfidence->code" not in bufr.data
    # Negative subscripts
    bufr = BUFRMessage(open("./sample-data/temp.bufr"))
    assert "centre" in bufr.data[-1]
    assert np.all(bufr.data[-1]["centre"] == bufr.data[2]["centre"])
    assert "pressure" in bufr.data[-2]
    assert np.all(bufr.data[-2]["pressure"] == bufr.data[1]["pressure"])
    assert "blockNumber" in bufr.data[-3]
    assert bufr.data[-3]["blockNumber"] == bufr.data[0]["blockNumber"]
    with pt.raises(IndexError):
        bufr.data[-4]


def test_data_get_count():
    bufr = BUFRMessage(open("./sample-data/synop.bufr"))
    with pt.raises(KeyValueNotFoundError):
        bufr.data.get_count("edition")
    assert bufr.get_count("year") == 1
    assert bufr.get_count("year->code") == 1
    assert bufr.get_count("year->percentConfidence") == 1
    assert bufr.get_count("year->percentConfidence->code") == 1
    assert bufr.get_count("#1#year") == 1
    assert bufr.get_count("#1#year->code") == 1
    assert bufr.get_count("#1#year->percentConfidence") == 1
    assert bufr.get_count("#1#year->percentConfidence->code") == 1
    with pt.raises(KeyValueNotFoundError):
        bufr.get_count("#2#year")
    assert bufr.get_count("cloudAmount") == 5
    assert bufr.get_count("cloudAmount->code") == 5
    assert bufr.get_count("cloudAmount->percentConfidence") == 5
    assert bufr.get_count("cloudAmount->percentConfidence->code") == 5
    assert bufr.get_count("#1#cloudAmount") == 1
    assert bufr.get_count("#1#cloudAmount->code") == 1
    assert bufr.get_count("#1#cloudAmount->percentConfidence") == 1
    assert bufr.get_count("#1#cloudAmount->percentConfidence->code") == 1
    with pt.raises(KeyValueNotFoundError):
        bufr.get_count("#6#cloudAmount")
    bufr = BUFRMessage(open("./sample-data/terra-modis-aerosol.bufr"))
    assert bufr.get_count("opticalDepth") == 10
    assert bufr.data[0].get_count("opticalDepth") == 2
    assert bufr.data[1].get_count("opticalDepth") == 8
    assert bufr.data.get_count("#1#opticalDepth") == 1
    assert bufr.data[0].get_count("#1#opticalDepth") == 1
    assert bufr.data[1].get_count("#1#opticalDepth") == 1
    with pt.raises(KeyValueNotFoundError):
        bufr.data.get_count("#11#opticalDepth")
    with pt.raises(KeyValueNotFoundError):
        bufr.data[0].get_count("#3#opticalDepth")
    with pt.raises(KeyValueNotFoundError):
        bufr.data[1].get_count("#9#opticalDepth")


def test_data_is_missing():
    bufr = BUFRMessage("BUFR3_local_satellite")
    bufr["numberOfSubsets"] = 3
    with pt.raises(KeyValueNotFoundError):
        bufr.data.is_missing("edition")
    assert bufr.is_missing("pressure")
    assert bufr.is_missing("#1#pressure")
    assert bufr.is_missing("#2#pressure")
    array = bufr["pressure"]
    assert np.all(array.mask)
    assert array.shape[0] > 1
    array[0, :] = 1013.25
    assert not np.all(array.mask)
    assert not bufr.is_missing("pressure")
    assert not bufr.is_missing("#1#pressure")
    assert bufr.is_missing("#2#pressure")
    bufr.pack()
    assert not bufr.is_missing("pressure")
    assert not bufr.is_missing("#1#pressure")
    assert bufr.is_missing("#2#pressure")


def test_data_set_missing():
    bufr = BUFRMessage("BUFR3_local_satellite")
    bufr["numberOfSubsets"] = 3
    with pt.raises(KeyValueNotFoundError):
        bufr.data.set_missing("edition")
    with pt.raises(ValueCannotBeMissingError):
        bufr.set_missing("pressure->units")
    bufr["pressure"] = 1013.25
    bufr.pack()
    bufr.set_missing("pressure")
    array = bufr["pressure"]
    assert np.all(array.mask)
    bufr["pressure"] = 1013.25
    bufr.pack()
    bufr.set_missing("#1#pressure")
    array = bufr["pressure"]
    assert np.all(array[0, :].mask)
    assert not np.any(array[1:, :].mask)
    bufr.set_missing("#2#pressure")
    assert np.all(array[1, :].mask)
    assert not np.any(array[2:, :].mask)
    bufr.set_missing("pressure")
    assert np.all(array.mask)


def test_data_type_of_seconds():
    # With default, zero scale -> int
    bufr = BUFRMessage("BUFR3_local_satellite")
    bufr["numberOfSubsets"] = 3
    bufr["unexpandedDescriptors"] = [4006]
    assert bufr["second->scale"] == 0
    assert bufr["second"].dtype == np.dtype(int)
    assert bufr["second"].data[0] == CODES_MISSING_LONG
    assert bufr["second"][0] is np.ma.masked
    # With changed, non-zero scale -> float
    bufr = BUFRMessage("BUFR3_local_satellite")
    bufr["numberOfSubsets"] = 3
    bufr["unexpandedDescriptors"] = [202131, 4006, 202000]
    assert bufr["second->scale"] == 3
    assert bufr["second"].dtype == np.dtype(float)
    assert bufr["second"].data[0] == CODES_MISSING_DOUBLE
    assert bufr["second"][0] is np.ma.masked


def test_data_assumed_scalar_elements():
    bufr = BUFRMessage("BUFR3_local_satellite")
    bufr["numberOfSubsets"] = 3
    bufr["unexpandedDescriptors"] = [1007]
    bufr.data["satelliteIdentifier"] = [1, 2, 3]
    bytes = bufr.get_buffer()
    bufr = BUFRMessage(bytes)
    with change_behaviour() as behaviour:
        behaviour.assumed_scalar_elements.add("satelliteIdentifier")
        with pt.warns(UserWarning):
            assert np.all(bufr.data["satelliteIdentifier"] == [1, 2, 3])
        with change_behaviour() as behaviour:
            behaviour.on_assumed_scalar_element_invalid_size = "raise"
            bufr = BUFRMessage(bytes)
            with pt.raises(ValueError):
                assert np.all(bufr.data["satelliteIdentifier"] == [1, 2, 3])
        with change_behaviour() as behaviour:
            behaviour.on_assumed_scalar_element_invalid_size = "ignore"
            bufr = BUFRMessage(bytes)
            with warnings.catch_warnings(record=True) as ws:
                assert np.all(bufr.data["satelliteIdentifier"] == [1, 2, 3])
                assert not ws
    with change_behaviour() as behaviour:
        behaviour.assumed_scalar_elements.add("satelliteIdentifier")
        # Single rank
        bufr = BUFRMessage("BUFR3_local_satellite")
        bufr["numberOfSubsets"] = 3
        bufr["unexpandedDescriptors"] = [1007]
        assert bufr.data["satelliteIdentifier"] is np.ma.masked
        assert bufr.data["#1#satelliteIdentifier"] is np.ma.masked
        bufr.data["satelliteIdentifier"] = 1
        bufr.data["#1#satelliteIdentifier"] = 1
        bufr.pack()
        assert bufr.data["satelliteIdentifier"] == 1
        with pt.raises(ValueError):
            bufr.data["satelliteIdentifier"] = [1, 2, 3]
        # Multiple ranks
        bufr = BUFRMessage("BUFR3_local_satellite")
        bufr["numberOfSubsets"] = 3
        bufr["unexpandedDescriptors"] = [1007, 1007]
        value = bufr.data["satelliteIdentifier"]
        assert isinstance(value, np.ndarray)
        assert value.shape == (2,)
        assert value.all() is np.ma.masked
        value1 = bufr.data["#1#satelliteIdentifier"]
        assert value1 is np.ma.masked
        bufr.data["satelliteIdentifier"] = [0, 0]
        assert np.all(bufr.data["satelliteIdentifier"] == 0)
        bufr.data["#1#satelliteIdentifier"] = 1
        bufr.data["#2#satelliteIdentifier"] = 2
        assert np.all(bufr.data["satelliteIdentifier"] == [1, 2])


def test_data_assignment_with_non_native_types():
    bufr = BUFRMessage("BUFR3_local_satellite")
    bufr["satelliteIdentifier"] = np.int8(0)
    bufr["satelliteIdentifier"] = np.int16(0)
    bufr["satelliteIdentifier"] = np.int32(0)
    bufr["latitude"] = np.float32(0)



# Copyright 2022- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

# flake8: noqa: F405
#   ruff: noqa: F403

from .common import *


class RaggedArray:
    @staticmethod
    def empty(ndim=1):
        if ndim > 0:
            data = []
            for _ in range(ndim - 1):
                data = [data]
        else:
            data = None
        return RaggedArray(data, ndim)

    def __init__(self, data, ndim=None):
        self.data = data
        if ndim is None:
            self.ndim = 0
            while True:
                try:
                    data = data[0]
                except (IndexError, TypeError):
                    break
                else:
                    self.ndim += 1
        else:
            self.ndim = ndim

    def __repr__(self):
        return f"RaggedArray({self.data})"

    def __getitem__(self, index):
        try:
            value = self.data[index[0]]
        except (IndexError, TypeError):
            try:
                value = self.data[index]
            except TypeError:
                if index in ((), None):
                    value = self.data
                else:
                    raise IndexError(f"{index}")
        else:
            for i in range(1, len(index)):
                value = value[index[i]]
        return value

    def __setitem__(self, index, value):
        data = self.data
        try:
            for i in range(0, len(index) - 1):
                data = data[index[i]]
        except (IndexError, TypeError):
            try:
                self.data[index] = value
            except TypeError:
                if index in ((), None):
                    self.data = value
                else:
                    raise IndexError(f"{index}")
        else:
            data[index[-1]] = value

    def __bool__(self):
        if isinstance(self.data, list):

            def recurse(data):
                if not isinstance(data, list):
                    return True
                elif len(data) == 0:
                    return False
                elif len(data) > 1:
                    return True
                else:
                    return recurse(data[0])

            return recurse(self.data)
        else:
            return bool(self.data)

    def insert(self, index, value):
        data = self.data
        try:
            for i in range(0, len(index) - 1):
                data = data[index[i]]
            data.insert(index[-1], value)
        except (IndexError, TypeError, AttributeError):
            try:
                data.insert(index, value)
            except (IndexError, TypeError):
                if index[i] == len(data):
                    data.append([])
                    data = data[index[i]]
                    data.insert(index[-1], value)
                else:
                    raise IndexError(f"{index}")
            except AttributeError:
                if index in ((), None):
                    self.data = value
                else:
                    raise IndexError(f"{index}")


class SingletonDict:  # TODO: UserDict?
    """A dict which returns the same value, no matter the key.

    This is a helper class to optimize storage of key counts in leaf nodes
    where the counts remain the same for every replication. By using this class
    instead of the build-in dict, we save space without having to special-case
    access to the leaf nodes' counts in the rest of the code.
    """

    def __init__(self):
        self._has_value = False

    def __getitem__(self, key):
        if self._has_value:
            return self._value
        else:
            raise KeyError(key)

    def __setitem__(self, key, value):
        # TODO: allow to set/update value only once
        self._value = value
        self._has_value = True


def get_datetime(
    view: Union["View", abc.Mapping],
    rank: Optional[Union[int, slice]] = None,
    prefix: str = "",
    year: Optional[int] = None,
    month: Optional[int] = None,
) -> NDArray:
    """
    Returns an array of type `datetime64` derived from datetime-related keys/values.
    """
    date = get_date(view, rank, prefix, year, month)
    time = get_time(view, rank, prefix)
    if date is np.ma.masked or time is np.ma.masked:
        datetime = np.ma.masked
    else:
        datetime = date + time
    return datetime


def get_date(
    view: Union["View", abc.Mapping],
    rank: Optional[Union[int, slice]] = None,
    prefix: str = "",
    year: Optional[int] = None,
    month: Optional[int] = None,
) -> NDArray:
    """
    Returns a `datetime64` array derived from date-related keys/values.
    """
    names = ("year", "month", "day")
    values = (year, month, None)
    dtypes = ("M8[Y]", "m8[M]", "m8[D]")
    offsets = (1970, 1, 1)
    parts = []
    for name, value, dtype, offset in zip(names, values, dtypes, offsets):
        if prefix:
            name = name.capitalize()
        key = Key(prefix + name, rank if isinstance(rank, int) else None)
        value = view[key.string] if value is None else value
        part = ensure_masked_value(value, dtype)
        parts.append(part - offset)
    for i, part in enumerate(parts):
        if part is np.ma.masked:
            date = np.ma.masked
            break
        if isinstance(rank, slice):
            slice_ = decrement(rank)
            parts[i] = part[slice_]
    else:
        date = sum(parts)
    return date


def get_time(
    view: Union["View", abc.Mapping],
    rank: Optional[Union[int, slice]] = None,
    prefix: str = "",
) -> NDArray:
    """
    Returns a `timedelta64` array derived from time-related keys/values.
    """
    names = ("hour", "minute", "second")
    dtypes = ("m8[h]", "m8[m]", "m8[s]")
    has_seconds = True
    parts = []
    for name, dtype in zip(names, dtypes):
        if prefix:
            name = name.capitalize()
        key = Key(prefix + name, rank if isinstance(rank, int) else None)
        try:
            value = view[key.string]
        except (NotFoundError, KeyError) as error:
            if name in ("second", "Second"):
                has_seconds = False
                break
            else:
                raise error
        if name == "second":
            if value is np.ma.masked:
                part = value
            else:
                array = ensure_masked_array(value)
                if array.dtype == np.dtype("f8"):
                    dtype = "m8[ms]"
                    array.fill_value = missing_of(int)
                    array.data[array.mask] = array.fill_value
                    array = (array * 1000).astype(dtype)
                    array.fill_value = np.array("NaT", dtype)
                elif array.dtype == np.dtype("i8"):
                    array = array.view(dtype)
                else:
                    array = cast(MaskedArray, array.astype(dtype))
                if array.dtype == np.dtype("m8[s]"):
                    max_value = np.array(60 - 1, array.dtype)
                elif array.dtype == np.dtype("m8[ms]"):
                    max_value = np.array(60 * 1000 - 1, array.dtype)
                else:
                    assert False
                array[:] = np.ma.where(array > max_value, max_value, array)
                if isinstance(value, abc.Iterable):
                    part = array
                else:
                    part = array[0]
        else:
            part = ensure_masked_value(value, dtype)
        parts.append(part)
    if not prefix:
        key = Key(
            "secondsWithinAMinuteMicrosecond", rank if isinstance(rank, int) else None
        )
        try:
            value = view[key.string]
        except (NotFoundError, KeyError):
            pass
        else:
            array = ensure_masked_array(value)
            useconds = np.ma.empty(array.shape, dtype="m8[us]")
            useconds.fill_value = np.array("NaT", dtype="m8[us]")
            useconds[:] = array * 1000000
            max_useconds = np.array(60 * 1000000 - 1, dtype="m8[us]")  # [1]
            useconds[:] = np.ma.where(useconds > max_useconds, max_useconds, useconds)
            if isinstance(value, abc.Iterable):
                part = useconds
            else:
                part = useconds[0]
            if has_seconds:
                parts.pop()  # [2]
            parts.append(part)
    for i, part in enumerate(parts):
        if part is np.ma.masked:
            time = np.ma.masked
            break
        if isinstance(rank, slice):
            slice_ = decrement(rank)
            parts[i] = part[slice_]
    else:
        time = sum(parts)
    return time

    # [1] Note that 'seconds...Microsecond' is a floating-point number where the
    #     whole part is seconds and the decimal part is microseconds.
    #
    # [2] Because of [1], if the message contains both keys, the 'second' and the
    #    'seconds...Microsecond', use the latter. (Yes, such messages do exist!)


def set_datetime(
    view: Union["View", abc.MutableMapping],
    value: Union[DateLike, np.ndarray],
    rank: Optional[int] = None,
    prefix: str = "",
) -> None:
    set_date(view, value, rank, prefix)
    set_time(view, value, rank, prefix)


def set_date(
    view: Union["View", abc.MutableMapping],
    value: Union[DateLike, np.ndarray],
    rank: Optional[int] = None,
    prefix: str = "",
) -> None:
    names = ["year", "month", "day"]
    keys = []
    for i, name in enumerate(names):
        if prefix:
            name = prefix + name.capitalize()
        keys.append(Key(name, rank).string)
    year, month, day = keys
    if isinstance(value, np.ndarray):
        view[year] = value.astype("M8[Y]").astype(int) + 1970
        view[month] = value.astype("M8[M]").astype(int) % 12 + 1
        view[day] = (value.astype("M8[D]") - value.astype("M8[M]")).astype(int) + 1
    else:
        date = ensure_date(value)
        view[year] = date.year
        view[month] = date.month
        view[day] = date.day


def set_time(
    view: Union["View", abc.MutableMapping],
    value: Union[TimeLike, np.ndarray],
    rank: Optional[int] = None,
    prefix: str = "",
) -> None:
    keys = ["hour", "minute", "second"]
    for i, name in enumerate(keys):
        if prefix:
            name = prefix + name.capitalize()
        keys[i] = Key(name, rank).string
    hour, minute, second = keys
    if isinstance(value, np.ndarray):
        delta = (value - value.astype("M8[D]")).astype("timedelta64[s]").view(int)
        view[hour] = delta // 3600
        view[minute] = (delta % 3600) // 60
        try:
            view[second] = delta % 60
        except (NotFoundError, KeyError):
            pass
    else:
        time = ensure_time(value)
        view[hour] = time.hour
        view[minute] = time.minute
        try:
            view[second] = time.second
        except (NotFoundError, KeyError):
            pass


def ensure_date(value: DateLike) -> dt.date:
    """Takes a date-like value and converts it to `dt.date`."""
    date: dt.date
    if isinstance(value, dt.date):
        date = value
    elif isinstance(value, dt.datetime):
        date = value.date()
    elif isinstance(value, np.datetime64):
        date = value.astype("M8[D]").item()
    elif isinstance(value, str):
        try:
            date = dt.datetime.fromisoformat(value).date()
        except ValueError:
            try:
                date = dt.datetime.strptime(value, "%Y%m%d").date()
            except ValueError:
                raise ValueError(f"Cannot convert string '{value}' to `dt.date`")
    else:
        raise TypeError("Cannot convert value of type `%s` to `dt.time`" % type(value))
    return date


def ensure_time(value: TimeLike) -> dt.time:
    """Takes a time-like value and converts it to `dt.time`."""
    if isinstance(value, dt.time):
        time = value
    elif isinstance(value, dt.datetime):
        time = value.time()
    elif isinstance(value, dt.timedelta):
        time = (dt.datetime.min + value).time()
    elif isinstance(value, np.datetime64):
        time = value.astype("M8[s]").item().time()
    elif isinstance(value, np.timedelta64):
        total_seconds = value.astype("m8[s]").astype(int)
        hours = total_seconds // 3600
        minutes = (total_seconds % 3600) // 60
        seconds = total_seconds % 60
        time = dt.time(hours, minutes, seconds)
    elif isinstance(value, str):
        for format in ["%H:%M:%S", "%H:%M", "%H%M%S", "%H%M"]:
            try:
                time = dt.datetime.strptime(value, format).time()
            except ValueError:
                continue
            else:
                break
        else:
            raise ValueError(f"Cannot convert string '{value}' to `dt.time`")
    else:
        raise TypeError("Cannot convert value of type `%s` to `dt.time`" % type(value))
    return time


def ensure_array(
    value: Any, dtype: Optional[DTypeLike] = None
) -> Union[NDArray, MaskedArray]:
    """Takes a value and converts it to `np.ndarray`."""
    if isinstance(value, np.ndarray):
        array = value
        if dtype:
            dtype = np.dtype(dtype)
            if dtype.type in (np.datetime64, np.timedelta64):
                if array.dtype.type == np.int64:
                    array = array.view(dtype)
                else:
                    array = array.astype(dtype)
            else:
                array = array.astype(dtype, copy=False)
    elif isinstance(value, abc.Iterable):
        if isinstance(value, str):
            array = np.array([value], dtype=dtype)
        else:
            array = np.array(value, dtype=dtype)
    else:
        array = np.array([value], dtype=dtype)
    return array


def ensure_masked_value(value: Any, dtype: Optional[DTypeLike]) -> MaskedArray:
    if value is np.ma.masked:
        return value
    else:
        array = ensure_masked_array(value, dtype)
        if isinstance(value, abc.Iterable):
            return array
        else:
            assert array.size == 1
            return array[0]


def ensure_masked_array(value: Any, dtype: Optional[DTypeLike] = None) -> MaskedArray:
    """Takes a value and converts it to `np.ma.MaskedArray`."""
    array = ensure_array(value, dtype)
    if isinstance(value, np.ma.core.MaskedConstant):
        masked_array = cast(np.ma.MaskedArray, array.ravel())  # [1]
    elif isinstance(array, np.ma.MaskedArray):
        masked_array = array
    else:
        try:
            missing = missing_of(array.dtype)
        except ValueError:
            masked_array = np.ma.array(array, mask=False, copy=False)  # [2]
        else:
            masked_array = np.ma.masked_equal(array, missing, copy=False)
    if dtype and np.dtype(dtype).type in (np.datetime64, np.timedelta64):
        masked_array.fill_value = np.array("NaT", dtype)
    return masked_array

    # [1] We are converting MaskedConstant (i.e., np.ma.masked) into 1d array
    #     so that we can do a slice assignment. This reduces number of code paths
    #     later on.
    #
    # [2] Assume no masked values for non-native types which cannot accomodate
    #     missing values, such as int16, float32, etc..


def decrement(value: Union[int, range, slice]) -> Union[int, range, slice]:
    """Decrements a value by one."""
    if isinstance(value, int):
        value -= 1
    elif isinstance(value, range):
        value = range(value.start - 1, value.stop - 1, value.step)
    elif isinstance(value, slice):
        start = None if value.start is None else value.start - 1
        stop = None if value.stop is None else value.stop - 1
        value = slice(start, stop, value.step)
    else:
        raise TypeError("Cannot decrement value of type `%s`" % type(value))
    return value


def flatten(items):
    """Flattens arbitrarily nested sequences.
    Examples:
        >>> list(flatten([1, (2, (3, 4), [[5]])]))
        [1, 2, 3, 4, 5]
        >>> list(flatten(1))
        [1]
    """
    try:
        for outer in items:
            try:
                for inner in flatten(outer):
                    yield inner
            except TypeError:
                yield outer
    except TypeError:
        yield items


MISSING_OF = {
    str: "",
    int: CODES_MISSING_LONG,
    float: CODES_MISSING_DOUBLE,
    np.int32: CODES_MISSING_LONG,
    np.int64: CODES_MISSING_LONG,
    np.float64: CODES_MISSING_DOUBLE,
    np.dtype("i4"): CODES_MISSING_LONG,
    np.dtype("i8"): CODES_MISSING_LONG,
    np.dtype("f8"): CODES_MISSING_DOUBLE,
}


def missing_of(obj: Any) -> Union[int, float, str]:
    """Returns corresponding missing value for the given object or type."""
    try:
        missing = MISSING_OF[obj]
    except (KeyError, TypeError):
        if isinstance(obj, np.ndarray):
            try:
                missing = missing_of(obj.dtype)
            except KeyError:
                missing = None
        elif isinstance(obj, np.dtype):
            if obj.type == np.str_:
                missing = ""
            elif obj.type in (np.datetime64, np.timedelta64):
                missing = np.array(CODES_MISSING_LONG, dtype=obj)
            else:
                missing = None
        elif hasattr(obj, "__iter__") and len(obj) > 0:
            missing = missing_of(type(obj[0]))
        elif not isinstance(obj, type):
            missing = missing_of(type(obj))
        else:
            missing = None
    if missing is None:
        raise ValueError("Object %s has no corresponding missing value" % obj)
    return missing



# Copyright 2022- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

# flake8: noqa: F403
# flake8: noqa: F405

import numpy as np
import pytest as pt

from eccodes.highlevel._bufr.helpers import *

# Test RaggedArray


def test_ragged_array_getset():
    a0 = RaggedArray(0)
    assert a0.ndim == 0
    assert a0[None] == 0
    assert a0[()] == 0
    a1 = RaggedArray([0, 1])
    assert a1.ndim == 1
    assert a1[0] == 0
    assert a1[(1,)] == 1
    a1[0] = -0
    a1[(1,)] = -1
    assert a1[0] == -0
    assert a1[(1,)] == -1
    a2 = RaggedArray([[0], [1, 2]])
    assert a2.ndim == 2
    assert a2[(0, 0)] == 0
    assert a2[(1, 0)] == 1
    assert a2[(1, 1)] == 2
    a2[(0, 0)] = -0
    a2[(1, 0)] = -1
    a2[(1, 1)] = -2
    assert a2[(0, 0)] == -0
    assert a2[(1, 0)] == -1
    assert a2[(1, 1)] == -2


def test_ragged_array_insert():
    a0 = RaggedArray.empty(ndim=0)
    assert a0.data is None
    assert a0.ndim == 0
    a0.insert((), 0)
    assert a0[()] == 0
    a1 = RaggedArray.empty(ndim=1)
    assert a1.data == []
    assert a1.ndim == 1
    a1.insert(0, 0)
    a1.insert((1,), 1)
    assert a1.data == [0, 1]
    a2 = RaggedArray.empty(ndim=2)
    assert a2.data == [[]]
    assert a2.ndim == 2
    a2.insert((0, 0), 0)
    assert a2.data == [[0]]
    a2.insert((1, 0), 1)
    a2.insert((1, 1), 2)
    assert a2.data == [[0], [1, 2]]


def test_ragged_array_bool():
    assert not bool(RaggedArray(0))
    assert not bool(RaggedArray([]))
    assert not bool(RaggedArray([[]]))
    assert bool(RaggedArray(1))
    assert bool(RaggedArray([0]))
    assert bool(RaggedArray([[0]]))


# Test get_datetime()


def test_get_datetime_with_no_seconds():
    data = dict(year=2000, month=1, day=2, hour=3, minute=4)
    datetime = get_datetime(data)
    assert not isinstance(datetime, np.ndarray)
    assert datetime == np.datetime64("2000-01-02T03:04")


def test_get_datetime_with_missing_seconds():
    data = dict(year=2000, month=1, day=2, hour=3, minute=4)
    data.update(second=CODES_MISSING_LONG)
    datetime = get_datetime(data)
    assert datetime is np.ma.masked
    data.update(second=np.ma.masked)
    datetime = get_datetime(data)
    assert datetime is np.ma.masked
    data.update(second=[5, CODES_MISSING_LONG])
    datetime = get_datetime(data)
    assert isinstance(datetime, np.ndarray)
    assert datetime.size == 2
    assert datetime[0] == np.datetime64("2000-01-02T03:04:05")
    assert datetime[1] is np.ma.masked
    assert np.ma.all(
        datetime == np.array(["2000-01-02T03:04:05", "NaT"], dtype="M8[s]")
    )


def test_get_datetime_with_decimal_seconds():
    data = dict(year=2000, month=1, day=2, hour=3, minute=4)
    data.update(second=[5.6, CODES_MISSING_DOUBLE])
    datetime = get_datetime(data)
    assert isinstance(datetime, np.ndarray)
    assert datetime.size == 2
    assert datetime[0] == np.datetime64("2000-01-02T03:04:05.6")
    assert datetime[1] is np.ma.masked
    assert np.ma.all(
        datetime == np.array(["2000-01-02T03:04:05.6", "NaT"], dtype="M8[ms]")
    )


def test_get_datetime_with_out_of_range_seconds():
    data = dict(year=2000, month=1, day=2, hour=3, minute=4)
    data.update(second=[5, 60])
    datetime = get_datetime(data)
    assert isinstance(datetime, np.ndarray)
    assert datetime.size == 2
    assert datetime[0] == np.datetime64("2000-01-02T03:04:05")
    assert datetime[1] == np.datetime64("2000-01-02T03:04:59")


def test_get_datetime_with_int32_seconds():
    data = dict(year=2000, month=1, day=2, hour=3, minute=4)
    data.update(second=np.array([5, 6], dtype="int32"))
    datetime = get_datetime(data)
    assert isinstance(datetime, np.ndarray)
    assert datetime.size == 2
    assert datetime[0] == np.datetime64("2000-01-02T03:04:05")
    assert datetime[1] == np.datetime64("2000-01-02T03:04:06")


def test_get_datetime_with_microseconds():
    data = dict(
        year=2000,
        month=1,
        day=2,
        hour=3,
        minute=4,
        secondsWithinAMinuteMicrosecond=5.6789,
    )
    datetime = get_datetime(data)
    assert not isinstance(datetime, np.ndarray)
    assert datetime == numpy.datetime64("2000-01-02T03:04:05.6789")


def test_get_datetime_with_seconds_and_microseconds():
    data = dict(
        year=2000,
        month=1,
        day=2,
        hour=3,
        minute=4,
        second=5,
        secondsWithinAMinuteMicrosecond=5.6789,
    )
    datetime = get_datetime(data)
    assert not isinstance(datetime, np.ndarray)
    assert datetime == numpy.datetime64("2000-01-02T03:04:05.6789")


def test_get_datetime_with_prefix():
    header = dict(rdbtimeDay=2, rdbtimeHour=3, rdbtimeMinute=4, rdbtimeSecond=5)
    with pt.raises(KeyError):
        datetime = get_datetime(header, prefix="rdbtime")
    datetime = get_datetime(header, prefix="rdbtime", year=2000, month=1)
    assert not isinstance(datetime, np.ndarray)
    assert datetime == numpy.datetime64("2000-01-02T03:04:05")


def test_get_datetime_with_rank():
    data = dict(year=[2000, 2001, 2002], month=[1, 2, 3], day=[2, 3, 4])
    data.update(hour=[3, 4, 5], minute=[4, 5, 6], second=[5, 6, 7])
    for name, values in list(data.items()):
        # data.pop(name)
        for rank, value in enumerate(values, 1):
            data[f"#{rank}#{name}"] = value
    # Number
    datetime = get_datetime(data, rank=2)
    assert not isinstance(datetime, np.ndarray)
    assert datetime == np.datetime64("2001-02-03T04:05:06")
    # Slice
    datetime = get_datetime(data, rank=slice(2, None))
    assert isinstance(datetime, np.ndarray)
    assert np.all(
        datetime
        == np.array(["2001-02-03T04:05:06", "2002-03-04T05:06:07"], dtype="M8[s]")
    )


# Test missing_of()


def test_missing_of():
    with pt.raises(ValueError):
        missing_of(np.int8)
    with pt.raises(ValueError):
        missing_of(np.int16)
    assert CODES_MISSING_LONG == missing_of(np.int32)
    assert CODES_MISSING_LONG == missing_of(np.int64)
    with pt.raises(ValueError):
        missing_of(np.dtype("i2"))
    with pt.raises(ValueError):
        missing_of(np.dtype("f4"))
    assert CODES_MISSING_LONG == missing_of(np.dtype("i4"))
    assert CODES_MISSING_LONG == missing_of(np.dtype("i8"))
    assert CODES_MISSING_LONG == missing_of(np.array([1]))
    assert CODES_MISSING_LONG == missing_of([1])
    assert CODES_MISSING_LONG == missing_of(1)
    with pt.raises(ValueError):
        missing_of(np.float16)
    with pt.raises(ValueError):
        missing_of(np.float32)
    assert CODES_MISSING_DOUBLE == missing_of(np.float64)
    assert CODES_MISSING_DOUBLE == missing_of(np.dtype("f8"))
    assert CODES_MISSING_DOUBLE == missing_of(np.array([1.0]))
    assert CODES_MISSING_DOUBLE == missing_of([1.0])
    assert CODES_MISSING_DOUBLE == missing_of(1.0)
    assert "" == missing_of(np.dtype("U"))
    assert "" == missing_of(np.array(["one"]))
    assert "" == missing_of(["one"])
    assert "" == missing_of("one")
    assert np.array(CODES_MISSING_LONG, "M8[D]") == missing_of(np.dtype("M8[D]"))
    assert np.array(CODES_MISSING_LONG, "M8[D]") == missing_of(np.array(1, "M8[D]"))
    assert np.array(CODES_MISSING_LONG, "m8[h]") == missing_of(np.dtype("m8[h]"))
    assert np.array(CODES_MISSING_LONG, "m8[h]") == missing_of(np.array(1, "m8[h]"))



# Copyright 2022- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

# flake8: noqa: F403
# flake8: noqa: F405
import sys

import numpy as np
import pytest as pt

import eccodes
from eccodes import *
from eccodes.highlevel._bufr import *

if sys.platform.startswith("win32"):
    pt.skip("Not applicable on Windows", allow_module_level=True)


@pt.fixture(autouse=True)
def fixture(monkeypatch):
    import os

    this_dir = os.path.dirname(__file__)
    monkeypatch.chdir(this_dir)  # always run tests from the tests directory


@pt.mark.skipif(
    eccodes.__version__ < "2.41.0",
    reason="requires eccodes 2.41.0; otherwise segfaults (see ECC-2024)",
)
def test_workaround_for_ECC_2015():
    old = BUFRMessage(open("./sample-data/hdob.bufr"))
    old.copy(subsets=[0])
    old.copy(subsets=[1])  # would fail without the workaround


def test_workaround_for_ECC_1624():
    # Consider the following message with operator 222000, and thus the 'centre'
    # key in the data section.
    bufr = BUFRMessage("BUFR3")
    bufr["numberOfSubsets"] = 1
    bufr["compressedData"] = 0
    bufr["unexpandedDescriptors"] = [
        307011,
        7006,
        10004,
        222000,
        101023,
        31031,
        1031,
        1032,
        101023,
        33007,
    ]

    # There should be only one element with key 'centre', and its
    # initial value should be missing.
    assert bufr.get_count("centre") == 1
    assert bufr["centre"] is np.ma.masked

    # Test that setting 'centre' value does the right thing, and doesn't interfere
    # with 'bufrHeaderCentre' from section 1.
    bufr["centre"] = 42
    bufr.pack()
    assert bufr["centre"] == 42
    assert bufr["bufrHeaderCentre"] == 98

    # Test that setting 'centre' on header raises an exception.
    with pt.raises(KeyValueNotFoundError):
        bufr.header["centre"]
    with pt.raises(KeyValueNotFoundError):
        bufr.header["centre"] = 42

    # Now consider a message *without* operator 222000, and thus *no* 'centre'
    # key in the data section.
    bufr = BUFRMessage("BUFR3")
    bufr["numberOfSubsets"] = 1
    bufr["compressedData"] = 0
    bufr["unexpandedDescriptors"] = [307011]

    # Test that setting 'centre' raises an exception.
    with pt.raises(KeyValueNotFoundError):
        bufr["centre"]
    with pt.raises(KeyValueNotFoundError):
        bufr["centre"] = 42


def test_workaround_for_ECC_2098():
    bufr = BUFRMessage("BUFR3")
    bufr["inputDelayedDescriptorReplicationFactor"] = 3
    bufr["unexpandedDescriptors"] = [
        104000,  # Delayed replication of 4 elements
        31001,  # Delayed replication factor
        204001,  # Operator: Add associated field
        31021,  # Associated field significance
        11001,  # Wind speed
        204000,  # Operator: Close associated field
    ]
    bufr["windDirection->associatedField"] = [0, 0, 0]  # OK
    bufr["windDirection->associatedField->associatedFieldSignificance"] = [
        1,
        1,
        1,
    ]  # Would raise ArrayTooSmallError when packing
    bufr.pack()


def test_workaround_for_ECC_2022():
    old = BUFRMessage(open("./sample-data/amv-goes-9.bufr", "rb"))
    new = old.copy()
    old.copy_to(new)
    assert old.data == new.data  # would fail without the workaround



# Copyright 2022- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

# flake8: noqa: F405
#   ruff: noqa: F403

from .common import *
from .helpers import get_datetime, set_datetime


class View:
    """A base class with some common methods: getters, setters, etc."""

    def __contains__(self, key: str) -> bool:
        raise NotImplementedError

    def __eq__(self, other: object) -> bool:
        raise NotImplementedError

    def __getitem__(self, subscript):
        raise NotImplementedError

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __setitem__(self, key: str, value) -> None:
        raise NotImplementedError

    def get_count(self, key: str) -> int:
        raise NotImplementedError

    def as_dict(self, ranked=False, depth=0, **kwds) -> Dict:
        raise NotImplementedError

    def get(self, key: str, default: Optional[ValueLike] = None) -> Optional[ValueLike]:
        """Returns value of `key`. If `key` is not defined, returns `default`.

        Note this method doesn't raise exception.
        """
        try:
            value = self[key]
        except NotFoundError:
            value = default
        return value

    def get_datetime(
        self,
        rank: Optional[Union[int, slice]] = None,
        prefix: str = "",
        year: Optional[int] = None,
        month: Optional[int] = None,
    ) -> NDArray:
        """
        Returns an array of type `datetime64` derived from datetime-related keys/values.

        The keys used are: 'year', 'month', 'day', 'hour', 'minute' and 'second'. If
        'second' is not present then the key 'secondsWithinAMinuteMicrosecond' is
        tried instead. The presence of keys 'second' and 'seconds...Microsecond' is
        optional.

        If `prefix` is specified, the datetime is derived from keys '{prefix}Year',
        '{prefix}Month', etc.

        Optionally, the year and the month can be overwritten/forced to a specific value.
        This can be useful, for instance, if there is no dedicated year or month key.
        For example, ECMWF's section 2 defines 'rdbtimeDay', 'rdbtimeHour', etc.,
        but not 'rdbtimeYear' or 'rdbtimeMonth'.
        """
        return get_datetime(self, rank, prefix, year, month)

    def items(self, ranked=False, **kwds) -> Iterator[Tuple[str, ValueLike]]:
        raise NotImplementedError

    def keys(self, ranked=False, **kwds) -> Iterator[str]:
        raise NotImplementedError

    def set(self, key: str, value: ValueLike) -> None:
        raise NotImplementedError

    def set_datetime(
        self,
        value: Union[DateLike, np.ndarray],
        rank: Optional[int] = None,
        prefix: str = "",
    ) -> None:
        """
        Sets datetime-related keys (i.e., 'year', 'month', etc.).

        If `prefix` is specified, use the keys '{prefix}Year', '{prefix}Month', etc.
        """
        set_datetime(self, value, rank, prefix)

    def update(self, *args, **kwargs) -> None:
        if len(args) == 1:
            arg = args[0]
            if isinstance(arg, View):
                for key, value in arg.items(skip="read_only"):
                    self[key] = value
            if isinstance(arg, abc.Mapping):
                for key, value in arg.items():
                    self[key] = value
            elif isinstance(arg, abc.Iterable):
                for key, value in arg:
                    self[key] = value
            else:
                raise TypeError(
                    "Expected a mapping or an iterable of key-value pairs; got %s"
                    % type(arg)
                )
        elif len(args) > 1:
            raise ValueError("Expected 1 positional argument; got %d" % len(args))
        else:
            self.update(kwargs)



# Copyright 2022- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime as dt
import enum
from collections import Counter, abc, defaultdict  # noqa
from contextlib import contextmanager
from copy import deepcopy
from dataclasses import dataclass, field, fields  # noqa
from enum import auto
from typing import (  # noqa
    Any,
    BinaryIO,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    cast,
)

import numpy
import numpy as np
from numpy.ma import MaskedArray  # noqa
from numpy.typing import DTypeLike, NDArray  # noqa

from eccodes.eccodes import KeyValueNotFoundError as NotFoundError
from eccodes.eccodes import *  # noqa

from .tables import Element


@dataclass
class Behaviour:
    assumed_scalar_elements: Set = field(default_factory=set)  # [1]
    on_assumed_scalar_element_invalid_size: str = "warn"  # [2]
    autorelease_handle: bool = True  # [3]
    update_header_from_data_before_packing: bool = True  # [4]


# [1] By default, values of data elements in uncompressed, multi-subset messages
#     are one-/two-dimensional arrays, even if there is only one common value for
#     all subsets. This is done in order to provide "stable" view of data elements
#     which doesn't depend on the encoding details. However, for situations where
#     the user can guarantee that an element, or a set of elements, will always
#     have only a single value (per rank), this option can be used to reduce
#     memory usage.
#
# [2] If an assumed scalar element is encoded with more than one value (per rank),
#     either print a warning ('warn'), raise an exception ('raise'), or just carry
#     on as if nothing hapened ('ignore').
#
# [3] Whether to release Coder's C handle automatically. The only reason you might
#     want to set this option to False is if you need to interoperate with the
#     low-level API, but you must know what you're doing.
#
# [4] This option controls whether to call Message's update_header_from_data()
#     method automatically before packing. This method sets some of the header
#     keys such as 'typicalYear', 'localMonth', 'localLatitude', etc.


DEFAULT_BEHAVIOUR = Behaviour(
    assumed_scalar_elements=set(
        [
            # 'originatorOfRetrievedAtmosphericConstituent',
            # 'satelliteChannelBandWidth',
            # 'satelliteChannelCentreFrequency',
            # 'satelliteIdentifier',
            # 'satelliteInstruments',
        ]
    )
)

current_behaviour = deepcopy(DEFAULT_BEHAVIOUR)


def get_behaviour():
    global current_behaviour  # noqa
    return deepcopy(current_behaviour)


def set_behaviour(new_behaviour):
    global current_behaviour  # noqa
    for f in fields(new_behaviour):
        setattr(current_behaviour, f.name, getattr(new_behaviour, f.name))


@contextmanager
def change_behaviour():
    global current_behaviour  # noqa
    saved_behaviour = get_behaviour()
    try:
        yield current_behaviour
    finally:
        set_behaviour(saved_behaviour)


MultiIndex = Tuple[int, ...]

DateLike = Union[dt.date, dt.datetime, np.datetime64, str]
TimeLike = Union[DateLike, dt.timedelta, np.timedelta64]
ValueLike = Union[bool, int, float, str, List, NDArray]


class Flags(enum.Flag):
    CODED = auto()
    COMPUTED = auto()
    FACTOR = auto()
    READ_ONLY = auto()
    BITMAP = auto()
    SCALAR = auto()


CODED = Flags.CODED
COMPUTED = Flags.COMPUTED
FACTOR = Flags.FACTOR
READ_ONLY = Flags.READ_ONLY
BITMAP = Flags.BITMAP
SCALAR = Flags.SCALAR

FLAGS = dict(
    coded=CODED, computed=COMPUTED, factor=FACTOR, read_only=READ_ONLY, bitmap=BITMAP
)

FlagsLike = Optional[Union[Flags, str, Sequence[str]]]


def ensure_flags(value: FlagsLike) -> Flags:
    if value is None:
        flags = Flags(0)
    elif isinstance(value, Flags):
        flags = value
    elif isinstance(name := value, str):
        try:
            flags = FLAGS[name]
        except KeyError:
            raise ValueError(f"'{name}' is not a valid flag name")
    elif isinstance(value, Sequence):
        names = list(value)
        flags = Flags(0)
        while names:
            name = names.pop()
            try:
                flags |= FLAGS[name]
            except KeyError:
                raise ValueError(f"'{name}' is not a valid flag name")
    else:
        raise TypeError("Invalid type: %s", type(value))
    return flags


@dataclass
class Key:
    name: str
    rank: Optional[int] = None
    primary: str = ""
    secondary: str = ""
    attribute: str = ""
    string: str = ""
    element: Optional[Element] = None
    flags: Flags = CODED

    @staticmethod
    def from_string(string: str):
        name, rank, primary, secondary, attribute = "", None, "", "", ""
        if string.startswith("/"):
            raise NotFoundError(
                f"{string}: Search keys with the '/<query>/<name>' syntax are not supported yet"
            )
        elif string.startswith("#"):
            second_hash = string.find("#", 2)
            if second_hash == -1:
                raise NotFoundError(f"'{string}': Missing second '#' character")
            rank = int(string[1:second_hash])
            if rank <= 0:
                raise NotFoundError(f"'{string}': Rank must be a positive number")
        else:
            second_hash = -1
        last_arrow = string.rfind("->", -11, -4)
        if last_arrow != -1:
            if string[last_arrow + 2 :] in (
                "code",
                "units",
                "reference",
                "scale",
                "width",
            ):
                attribute = string[last_arrow + 2 :]
            else:
                last_arrow = -1
        else:
            last_arrow = len(string)
        first_arrow = string.find("->", second_hash + 1, last_arrow)
        if first_arrow != -1:
            secondary = string[first_arrow + 2 : last_arrow]
        else:
            first_arrow = last_arrow
        primary = string[second_hash + 1 : first_arrow]
        name = string[second_hash + 1 : last_arrow]
        key = Key(name, rank, primary, secondary, attribute, string)
        return key

    def __post_init__(self):
        if not self.string:
            if self.rank:
                self.string = f"#{self.rank}#"
            self.string += self.name
            if self.attribute:
                self.string += f"->{self.attribute}"
        if not self.primary:
            self.primary = self.name


@dataclass
class Association:
    element: Element
    element_rank: int
    element_dtype: DTypeLike
    bitmap: NDArray
    bitmap_offset: int
    indices: Dict[str, NDArray]
    slices: Dict[str, slice] = field(init=False, default_factory=dict)  # [1]
    entries: Dict[str, "DataEntry"] = field(init=False, default_factory=dict)  # [2]

    # [1] Slices of elements sequence numbers which are within the scope of the bitmap.
    #
    # [2] A dict of secondary/associated entries keyed by the primary group name,.
    #     Only applies to those primary entries where at least one of the ranks is
    #     is set in the bitmap.

    def __post_init__(self) -> None:
        # Consider only those elements that are within the scope
        # of the bitmap.
        for key, indices in self.indices.items():
            bitmap_stop = self.bitmap.size + self.bitmap_offset
            start = numpy.searchsorted(indices >= self.bitmap_offset, True)
            stop = indices.size - numpy.searchsorted(indices[::-1] < bitmap_stop, True)
            self.slices[key] = slice(start, stop)

    def rank_mask(self, key: str):
        """Returns a mask indicating which ranks of the given primary key have
        an actual associated value.
        """
        slice = self.slices[key]
        indices = self.indices[key]  # all
        mask = numpy.repeat(
            False, len(indices)
        )  # the default for  ranks outside bitmap's scope
        indices = indices[slice]  # only within bitmap's scope
        indices = indices - self.bitmap_offset  # relative to bitmap
        mask[slice] = self.bitmap[indices]
        return mask

    def any_rank_set(self, key: str) -> bool:
        """Returns `True` if any rank of the given primary key has an actual
        associated value.
        """
        try:
            mask = self.rank_mask(key)
        except KeyError:  # key was defined *after* the bitmap association
            return False
        return bool(numpy.any(mask))


@dataclass
class DataEntry:
    name: str
    shape: Tuple = ()
    uniform_element: Optional[Element] = None
    elements: List[Element] = field(default_factory=list)
    array: Optional[NDArray] = None
    association: Optional[Association] = None
    primary: Optional["DataEntry"] = None
    flags: Flags = CODED

    @property
    def size(self) -> int:
        assert len(self.shape) == 2
        return self.shape[0] * self.shape[1]



# Copyright 2022- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import io
import sys
from pathlib import Path

import pytest

if sys.platform.startswith("win32"):
    pytest.skip("Not applicable on Windows", allow_module_level=True)


@pytest.fixture(autouse=True)
def fixture(monkeypatch):
    this_dir = Path(__file__).parent
    monkeypatch.syspath_prepend(this_dir.parent)
    # monkeypatch.chdir(this_dir.parent / 'data')
    monkeypatch.chdir(this_dir / "sample-data")


def compare(output: str, reference: Path):
    if not reference.exists():
        reference.write_text(output)
    try:
        assert output == reference.read_text()
    except AssertionError as error:
        reference.with_suffix(reference.suffix + ".bad").write_text(output)
        raise error
    else:
        reference.with_suffix(reference.suffix + ".bad").unlink(missing_ok=True)


inputs = [
    "acars.bufr",
    "ahi-himawari-8.bufr",
    "amsu-a-noaa-19.bufr",
    "amv-goes-9.bufr",
    "amv-insat-3d.bufr",
    "amv-meteosat-9.bufr",
    "amv-noaa-20.bufr",
    "aura-omi-ak.bufr",
    "buoy-drifting.bufr",
    "geos-abi-goes-16.bufr",
    "geos-mviri-meteosat-7.bufr",
    "hdob.bufr",
    "rwp.bufr",
    "rwp_jma.bufr",
    "saral-altika.bufr",
    "sral_sentinel_3a.bufr",
    "synop.bufr",
    "synop_multi_subset.bufr",
    "temp.bufr",
    "terra-modis-aerosol.bufr",
    "wave.bufr",
]


@pytest.mark.parametrize("input", inputs)
def test_attributes_example(input):
    from examples.attributes import run_example

    input = Path(input)
    stream = io.StringIO()
    run_example(input, output=stream)
    output = stream.getvalue()
    reference = input.with_suffix(".attributes")
    compare(output, reference)


@pytest.mark.parametrize("input", inputs)
def test_items_example(input):
    from examples.items import run_example

    input = Path(input)
    stream = io.StringIO()
    run_example(input, stream)
    output = stream.getvalue()
    reference = input.with_suffix(".items")
    compare(output, reference)
